# Copyright (c) 2014 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from testtools import matchers

from murano.dsl import serializer
from murano.tests.unit.dsl.foundation import object_model as om
from murano.tests.unit.dsl.foundation import test_case


# Tests for correctness of serialization of MuranoPL object tree
# back to Object Model

class TestResultsSerializer(test_case.DslTestCase):
    def setUp(self):
        super(TestResultsSerializer, self).setUp()
        self._class2 = om.Object('SampleClass2',
                                 class2Property='string2')
        self._class1 = om.Object(
            'SampleClass1', 'def',
            stringProperty='string1',
            arbitraryProperty={'a': [1, 2]},
            classProperty=self._class2)
        self._root_class = om.Object('ContractExamples', 'abc',
                                     sampleClass=self._class1)

        self._runner = self.new_runner(self._root_class)

    def _test_data_in_section(self, name, serialized):
        """Test that Model -> Load -> Serialize = Model

        Test that a section of Object Model has expected structure and
           property values that are equal to those originally loaded
           (e.g. that Model -> Load -> Serialize = Model)

           :param name: section name
           :param serialized: serialized Object Model
        """

        self.assertEqual('abc',
                         serialized[name]['?']['id'])
        self.assertEqual('ContractExamples/0.0.0@tests',
                         serialized['Objects']['?']['type'])
        self.assertIsInstance(serialized[name]['sampleClass'], dict)
        self.assertEqual('def',
                         serialized[name]['sampleClass']['?']['id'])
        self.assertEqual('SampleClass1/0.0.0@tests',
                         serialized[name]['sampleClass']['?']['type'])
        self.assertEqual('string1',
                         serialized[name]['sampleClass']['stringProperty'])

    def test_results_serialize(self):
        """Test that serialized contains same values and headers

        Test that serialized Object Model has both Objects and ObjectsCopy
        sections and they both contain the same property values and object
        headers. Note, that Objects section may contain additional designer
        metadata and information on available actions that is not needed in
        ObjectsCopy thus we cannot test for ObjectsCopy be strictly equal to
        Objects
        """

        serialized = self._runner.serialized_model
        self.assertIn('Objects', serialized)
        self.assertIn('ObjectsCopy', serialized)
        self._test_data_in_section('Objects', serialized)
        self._test_data_in_section('ObjectsCopy', serialized)

    def test_actions(self):
        """Test that information on actions can be invoked

        Test that information on actions can be invoked on each MuranoPL
        object are persisted into object header ('?' key) during serialization
        of Objects section of Object Model

        """
        serialized = self._runner.serialized_model
        actions = serialized['Objects']['?'].get('_actions')
        self.assertIsInstance(actions, dict)
        action_names = [action['name'] for action in actions.values()]
        self.assertIn('testAction', action_names)
        self.assertNotIn('notAction', action_names)
        self.assertIn('testRootMethod', action_names)
        action_meta = None
        for action in actions.values():
            self.assertIsInstance(action.get('enabled'), bool)
            self.assertIsInstance(action.get('name'), str)
            self.assertThat(
                action['name'],
                matchers.StartsWith('test'))
            if action['name'] == 'testActionMeta':
                action_meta = action
            else:
                self.assertEqual(action['title'], action['name'])
        self.assertIsNotNone(action_meta)
        self.assertEqual(action_meta['title'], "Title of the method")
        self.assertEqual(action_meta['description'],
                         "Description of the method")
        self.assertEqual(action_meta['helpText'], "HelpText of the method")

    def test_attribute_serialization(self):
        """Test that attributes produced by MuranoPL code are persisted

        Test that attributes produced by MuranoPL code are persisted in
        dedicated section of Object Model. Attributes are values that are
        stored in special key-value storage that is private to each class.
        Classes can store state data there. Attributes are persisted across
        deployment sessions but are not exposed via API (thus cannot be
        accessed by dashboard)
        """

        self._runner.on(self._class1).testAttributes('VALUE')
        serialized = self._runner.serialized_model
        self.assertIsInstance(serialized.get('Attributes'), list)
        self.assertThat(
            serialized['Attributes'],
            matchers.HasLength(1))
        self.assertEqual(
            [self._class1.id, 'SampleClass1', 'att1', 'VALUE'],
            serialized['Attributes'][0])

    def test_attribute_deserialization(self):
        """Test that attributes are available

        Test that attributes that are put into Attributes section of
        Object Model become available to appropriate MuranoPL classes
        """

        serialized = self._runner.serialized_model
        serialized['Attributes'].append(
            [self._class1.id, 'SampleClass1', 'att2', ' Snow'])
        runner2 = self.new_runner(serialized)
        self.assertEqual(
            'John Snow',
            runner2.on(self._class1).testAttributes('John'))

    def test_value_deserialization(self):
        """Test serialization of arbitrary values

        Test serialization of arbitrary values that can be returned
        from action methods
        """

        runner = self.new_runner(self._class2)
        result = runner.testMethod()
        self.assertEqual(
            {
                'key1': 'abc',
                'key2': ['a', 'b', 'c'],
                'key3': None,
                'key4': False,
                'key5': {'x': 'y'},
                'key6': [{'w': 'q'}]
            },
            serializer.serialize(result, runner.executor))



# Copyright (c) 2016 AT&T Corp
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from murano.packages import exceptions
import murano.tests.unit.base as test_base


class TestExceptions(test_base.MuranoTestCase):

    def test_package_class_load_error(self):
        class_name = 'test class name'
        message = 'test message'
        error = exceptions.PackageClassLoadError(class_name=class_name,
                                                 message=message)
        expected = 'Unable to load class "{0}" from package: {1}'\
                   .format(class_name, message)
        self.assertEqual(expected, error.args[0])

    def test_package_ui_load_error(self):
        messages = ['', 'test_message']
        for message in messages:
            error = exceptions.PackageUILoadError(message=message)
            expected = 'Unable to load ui definition from package'
            if message:
                expected += ': {0}'.format(message)
            self.assertEqual(expected, error.args[0])

    def test_package_format_error(self):
        messages = ['', 'test_message']
        for message in messages:
            error = exceptions.PackageFormatError(message=message)
            expected = 'Incorrect package format'
            if message:
                expected += ': {0}'.format(message)
            self.assertEqual(expected, error.args[0])



#    Copyright (c) 2014 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import re

from yaql.language import exceptions as yaql_exceptions
from yaql.language import expressions

from murano.dsl import constants
from murano.dsl import dsl_types
from murano.dsl import yaql_integration


EXPRESSION_REGEX = re.compile('^[\s\w\d.]*$')


class YaqlExpression(dsl_types.YaqlExpression):
    def __init__(self, expression, version):
        self._version = version
        if isinstance(expression, str):
            self._expression = str(expression)
            self._parsed_expression = yaql_integration.parse(
                self._expression, version)
            self._file_position = None
        elif isinstance(expression, YaqlExpression):
            self._expression = expression._expression
            self._parsed_expression = expression._parsed_expression
            self._file_position = expression._file_position
        elif isinstance(expression, expressions.Statement):
            self._expression = str(expression)
            self._parsed_expression = expression
            self._file_position = None
        else:
            raise TypeError('expression is not of supported types')

    @property
    def expression(self):
        return self._expression

    @property
    def version(self):
        return self._version

    @property
    def source_file_position(self):
        return self._file_position

    @source_file_position.setter
    def source_file_position(self, value):
        self._file_position = value

    def __repr__(self):
        return 'YAQL(%s)' % self._expression

    def __str__(self):
        return self._expression

    @staticmethod
    def is_expression(expression, version):
        if not isinstance(expression, str):
            return False
        if EXPRESSION_REGEX.match(expression):
            return False
        try:
            yaql_integration.parse(expression, version)
            return True
        except yaql_exceptions.YaqlParsingException:
            return False

    def __call__(self, context):
        if context:
            context[constants.CTX_CURRENT_INSTRUCTION] = self
        return self._parsed_expression.evaluate(context=context)



# Copyright (c) 2015 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from yaql.language import specs
from yaql.language import yaqltypes

from murano.common import engine
from murano.dsl import constants
from murano.dsl import dsl
from murano.dsl import helpers
from murano.dsl import yaql_integration


class MockContextManager(engine.ContextManager):

    def __init__(self):
        # { object_id: [mock_function_definitions] }
        self.obj_mock_ctx = {}

        # { class_name: [mock_function_definitions] }
        self.class_mock_ctx = {}

    def _create_new_ctx(self, mock_funcs):
        mock_context = yaql_integration.create_empty_context()
        for mock_func in mock_funcs:
            mock_context.register_function(mock_func)
        return mock_context

    def _create_new_ctx_for_class(self, name):
        new_context = None
        if name in self.class_mock_ctx:
            new_context = self._create_new_ctx(self.class_mock_ctx[name])
        return new_context

    def _create_new_ctx_for_obj(self, obj_id):
        new_context = None
        if obj_id in self.obj_mock_ctx:
            new_context = self._create_new_ctx(self.obj_mock_ctx[obj_id])
        return new_context

    def create_type_context(self, murano_type):
        original_context = super(
            MockContextManager, self).create_type_context(murano_type)

        mock_context = self._create_new_ctx_for_class(murano_type.name)
        if mock_context:
            result_context = helpers.link_contexts(
                original_context, mock_context).create_child_context()
        else:
            result_context = original_context
        return result_context

    def create_object_context(self, murano_obj):
        original_context = super(
            MockContextManager, self).create_object_context(murano_obj)

        mock_context = self._create_new_ctx_for_obj(murano_obj.type.name)
        if mock_context:
            result_context = helpers.link_contexts(
                original_context, mock_context).create_child_context()
        else:
            result_context = original_context
        return result_context

    def create_root_context(self, runtime_version):
        root_context = super(MockContextManager, self).create_root_context(
            runtime_version)
        constext = root_context.create_child_context()
        constext.register_function(inject_method_with_str, name='inject')
        constext.register_function(inject_method_with_yaql_expr,
                                   name='inject')
        constext.register_function(with_original)
        return constext


@specs.parameter('kwargs', yaqltypes.Lambda(with_context=True))
def with_original(context, **kwargs):
    new_context = context.create_child_context()

    original_context = context[constants.CTX_ORIGINAL_CONTEXT]
    for k, v in kwargs.items():
        new_context['$' + k] = v(original_context)
    return new_context


@specs.parameter(
    'target',
    yaqltypes.AnyOf(dsl.MuranoTypeParameter(), dsl.MuranoObjectParameter()))
@specs.parameter('target_method', yaqltypes.String())
@specs.parameter('mock_object', dsl.MuranoObjectParameter())
@specs.parameter('mock_name', yaqltypes.String())
def inject_method_with_str(context, target, target_method,
                           mock_object, mock_name):
    ctx_manager = helpers.get_executor().context_manager

    current_class = helpers.get_type(context)
    mock_func = current_class.find_single_method(mock_name)
    original_class = target.type

    original_function = original_class.find_single_method(target_method)
    result_fd = original_function.instance_stub.clone()

    def payload_adapter(__context, __sender, *args, **kwargs):
        return mock_func.invoke(
            mock_object, args, kwargs, __context, True)

    result_fd.payload = payload_adapter
    existing_mocks = ctx_manager.class_mock_ctx.setdefault(
        original_class.name, [])
    existing_mocks.append(result_fd)


@specs.parameter(
    'target',
    yaqltypes.AnyOf(dsl.MuranoTypeParameter(), dsl.MuranoObjectParameter()))
@specs.parameter('target_method', yaqltypes.String())
@specs.parameter('expr', yaqltypes.Lambda(with_context=True))
def inject_method_with_yaql_expr(context, target, target_method, expr):
    ctx_manager = helpers.get_executor().context_manager
    original_class = target.type

    original_function = original_class.find_single_method(target_method)
    result_fd = original_function.instance_stub.clone()

    def payload_adapter(__super, __context, __sender, *args, **kwargs):
        new_context = context.create_child_context()
        new_context[constants.CTX_ORIGINAL_CONTEXT] = __context
        mock_obj = context[constants.CTX_THIS]
        new_context.register_function(lambda: __super(*args, **kwargs),
                                      name='originalMethod')
        return expr(new_context, mock_obj, *args, **kwargs)

    result_fd.payload = payload_adapter
    result_fd.insert_parameter('__super', yaqltypes.Super())
    existing_mocks = ctx_manager.class_mock_ctx.setdefault(
        original_class.name, [])
    existing_mocks.append(result_fd)



#    Copyright (c) 2013 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json

from murano.db import models as m
from murano.db import session as db_session


class Statistics(object):

    @staticmethod
    def get_all():
        db = db_session.get_session()
        stats = db.query(m.ApiStats).all()
        return stats

    @staticmethod
    def get_stats_by_id(stats_id):
        db = db_session.get_session()
        stats = db.query(m.ApiStats).get(stats_id)
        return stats

    @staticmethod
    def get_stats_by_host(host):
        db = db_session.get_session()
        stats = db.query(m.ApiStats).filter(m.ApiStats.host == host).first()
        return stats

    @staticmethod
    def create(host, request_count, error_count,
               average_response_time, requests_per_tenant, cpu_count,
               cpu_percent):
        stats = m.ApiStats()
        stats.host = host
        stats.request_count = request_count
        stats.error_count = error_count
        stats.average_response_time = average_response_time
        stats.requests_per_tenant = json.dumps(requests_per_tenant)
        stats.requests_per_second = 0.0
        stats.errors_per_second = 0.0
        stats.cpu_count = cpu_count
        stats.cpu_percent = cpu_percent

        db = db_session.get_session()
        stats.save(db)

    @staticmethod
    def update(host, stats):
        db = db_session.get_session()
        stats.save(db)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from unittest import mock

from murano.dsl import attribute_store
from murano.dsl import dsl
from murano.dsl import dsl_types
from murano.tests.unit.dsl.foundation import test_case


class TestAttributeStore(test_case.DslTestCase):
    def setUp(self):
        super(TestAttributeStore, self).setUp()
        self.attribute_store = attribute_store.AttributeStore()
        self.fake_object = mock.MagicMock(object_id=mock.sentinel.oid)
        self.tagged_obj = dsl.MuranoObjectInterface(self.fake_object)
        self.owner_type = dsl_types.MuranoTypeReference(self.fake_object)
        self.owner_type.type.name = mock.sentinel.typename
        self.name = 'foobar'

    def test_get_attribute_key(self):
        oid, typename = mock.sentinel.oid, mock.sentinel.typename

        key = self.attribute_store._get_attribute_key(
            self.tagged_obj, self.owner_type, self.name)
        expected_key = (oid, (typename, 'foobar'))

        self.assertEqual(expected_key, key)

    @mock.patch.object(attribute_store.AttributeStore, '_get_attribute_key',
                       return_value=(mock.sentinel.key1, mock.sentinel.key2))
    def test_get(self, mock_get_attr_key):
        key1, key2 = mock.sentinel.key1, mock.sentinel.key2
        get_val = mock.sentinel.get_val

        self.attribute_store._attributes = mock.MagicMock()
        self.attribute_store._attributes[key1].get.return_value = get_val
        val = self.attribute_store.get(
            self.tagged_obj, self.owner_type, self.name)

        mock_get_attr_key.assert_called_with(
            self.tagged_obj, self.owner_type, self.name)
        self.attribute_store._attributes[key1].get.assert_called_with(key2)
        self.assertEqual(get_val, val)

    @mock.patch.object(attribute_store.AttributeStore, '_get_attribute_key',
                       return_value=(mock.sentinel.key1, mock.sentinel.key2))
    def test_set_object_if(self, mock_get_attr_key):
        val = dsl.MuranoObjectInterface(self.fake_object)
        self.attribute_store._attributes = mock.MagicMock()
        self.attribute_store.set(
            self.tagged_obj, self.owner_type, self.name, val)

    @mock.patch.object(attribute_store.AttributeStore, '_get_attribute_key',
                       return_value=(mock.sentinel.key1, mock.sentinel.key2))
    def test_set_object(self, mock_get_attr_key):
        key1, key2 = mock.sentinel.key1, mock.sentinel.key2

        val = dsl_types.MuranoObject()
        val.object_id = mock.sentinel.oid
        self.attribute_store.set(
            self.tagged_obj, self.owner_type, self.name, val)
        self.assertEqual(self.attribute_store._attributes[key1][key2],
                         mock.sentinel.oid)

    @mock.patch.object(attribute_store.AttributeStore, '_get_attribute_key',
                       return_value=(mock.sentinel.key1, mock.sentinel.key2))
    def test_set_none(self, mock_get_attr_key):
        key1, key2 = mock.sentinel.key1, mock.sentinel.key2

        val = None
        self.attribute_store._attributes = mock.MagicMock()
        self.attribute_store.set(
            self.tagged_obj, self.owner_type, self.name, val)

        self.attribute_store._attributes[key1].pop.assert_called_with(
            key2, None)

    def test_serialize(self):
        known_objects = ['obj1', 'obj3']
        self.attribute_store._attributes = {
            'obj1': {
                ('foo', 'obj11'): 11,
                ('bar', 'obj12'): 12
            },
            'obj2': {
                ('baz', 'obj21'): 21
            },
            'obj3': {
                ('foobar', 'obj31'): 31
            }
        }
        val = self.attribute_store.serialize(known_objects)
        expected = [
            ['obj1', 'foo', 'obj11', 11],
            ['obj1', 'bar', 'obj12', 12],
            ['obj3', 'foobar', 'obj31', 31]]

        self.assertEqual(sorted(expected), sorted(val))

    def test_load(self):
        data = [
            ['a', 'b', 'c', 'd'],
            ['a', 'f', 'g', None],
            ['b', 'i', 'j', 'k']
        ]

        self.attribute_store.load(data)
        expected = {'a': {('b', 'c'): 'd'},
                    'b': {('i', 'j'): 'k'}}
        self.assertEqual(expected, self.attribute_store._attributes)

    def test_forget_object_if(self):
        obj = dsl.MuranoObjectInterface(mock.MagicMock(object_id='bar'))
        self.attribute_store._attributes = {'foo': 42, 'bar': 43}
        self.attribute_store.forget_object(obj)
        self.assertEqual({'foo': 42}, self.attribute_store._attributes)

    def test_forget_object(self):
        obj = dsl_types.MuranoObject()
        obj.object_id = 'foo'
        self.attribute_store._attributes = {'foo': 42, 'bar': 43}
        self.attribute_store.forget_object(obj)
        self.assertEqual({'bar': 43}, self.attribute_store._attributes)



# Copyright (c) 2014 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from murano.dsl import exceptions
from murano.tests.unit.dsl.foundation import object_model as om
from murano.tests.unit.dsl.foundation import test_case


class TestPropertyAccess(test_case.DslTestCase):
    def setUp(self):
        super(TestPropertyAccess, self).setUp()
        self._multi_derived = om.Object(
            'DerivedFrom2Classes',
            rootProperty='ROOT',
            ambiguousProperty=321)
        model = om.Object(
            'SampleClass3',
            multiClassProperty=self._multi_derived
        )
        self._runner = self.new_runner(model)

    def test_private_property_access(self):
        self.assertIsNone(self._runner.testPrivateProperty())
        self.assertEqual(
            ['CommonParent', 'ParentClass1', 'SampleClass3'],
            self.traces)

    def test_private_property_access_failure(self):
        e = self.assertRaises(
            exceptions.UninitializedPropertyAccessError,
            self._runner.testUninitializedPrivatePropertyAccess)
        self.assertEqual(
            'Access to uninitialized property "privateName" '
            'in class "SampleClass3" is forbidden', str(e))

    def test_read_of_private_property_of_other_class(self):
        e = self.assertRaises(
            exceptions.PropertyAccessError,
            self._runner.testReadOfPrivatePropertyOfOtherClass)
        self.assertEqual(
            'Property "privateProperty" in class "DerivedFrom2Classes" '
            'cannot be read', str(e))
        self.assertEqual(['accessing property'], self.traces)

    def test_write_of_private_property_of_other_class(self):
        e = self.assertRaises(
            exceptions.PropertyAccessError,
            self._runner.testWriteOfPrivatePropertyOfOtherClass)
        self.assertEqual(
            'Property "privateProperty" in class "DerivedFrom2Classes" '
            'cannot be written', str(e))

    def test_access_ambiguous_property_with_resolver(self):
        self.assertEqual(
            '321',
            self._runner.on(self._multi_derived).
            testAccessAmbiguousPropertyWithResolver())

    def test_property_merge(self):
        self.assertEqual(
            '555',
            self._runner.on(self._multi_derived).
            testPropertyMerge())
        self.assertEqual(
            ['321', '555', '555', '555', '555'],
            self.traces)

    def test_property_usage(self):
        e = self.assertRaises(
            exceptions.NoWriteAccessError,
            self._runner.on(self._multi_derived).
            testModifyUsageTestProperty1)
        self.assertEqual(
            'Property "usageTestProperty1" is immutable to the caller',
            str(e))
        self.assertRaises(
            exceptions.NoWriteAccessError,
            self._runner.on(self._multi_derived).
            testModifyUsageTestProperty2)
        self.assertEqual(
            33,
            self._runner.on(self._multi_derived).
            testModifyUsageTestProperty3())
        self.assertEqual(
            44,
            self._runner.on(self._multi_derived).
            testModifyUsageTestProperty4())
        self.assertEqual(
            55,
            self._runner.on(self._multi_derived).
            testModifyUsageTestProperty5())
        self.assertRaises(
            exceptions.NoWriteAccessError,
            self._runner.on(self._multi_derived).
            testModifyUsageTestProperty6)
        self.assertEqual(
            77,
            self._runner.on(
                self._multi_derived).testModifyUsageTestProperty7())



# coding: utf-8
# Copyright (c) 2015 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from murano.dsl import dsl_exception
from murano.tests.unit.dsl.foundation import object_model as om
from murano.tests.unit.dsl.foundation import test_case


class TestUnicode(test_case.DslTestCase):
    def setUp(self):
        super(TestUnicode, self).setUp()
        self._runner = self.new_runner(om.Object('TestUnicode'))

    def test_literal(self):
        self.assertEqual(
            u"солнце ♥ φεγγάρι",
            self._runner.testLiteral())

    def test_expression(self):
        self.assertEqual(
            u"СОЛНЦЕ ♥ ΦΕΓΓΆΡΙ",
            self._runner.testExpression())

    def test_parameter(self):
        self.assertEqual(
            u"СОЛНЦЕ ♥ ΦΕΓΓΆΡΙ",
            self._runner.testParameter())

    def test_exception(self):
        x = self.assertRaises(
            dsl_exception.MuranoPlException,
            self._runner.testException)
        self.assertEqual(u"солнце ♥ φεγγάρι", x.message)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
from unittest import mock

from murano.dsl import helpers
from murano.dsl import session_local_storage
from murano.tests.unit.dsl.foundation import test_case


class FakeWithInit(session_local_storage._localbase):
    def __init__(self, *args, **kwargs):
        pass


class FakeNoInit(session_local_storage._localbase):
    pass


class TestLocalbase(test_case.DslTestCase):
    def test_new(self):
        lb = FakeWithInit.__new__(
            FakeWithInit, 42, 'foo', bar='baz')
        self.assertEqual(((42, 'foo',), {'bar': 'baz'}), lb._local__args)

    def test_new_bad(self):
        self.assertRaises(
            TypeError, FakeNoInit.__name__)


class TestPatch(test_case.DslTestCase):
    @mock.patch.object(helpers, 'get_execution_session',
                       return_value=mock.sentinel.session)
    def test_patch(self, mock_ges):
        fwi = FakeWithInit(foo='bar')
        session_local_storage._patch(fwi)
        self.assertEqual({}, fwi.__dict__)


class TestLocal(test_case.DslTestCase):
    class FakeLocal(session_local_storage._local):
        def __init__(self, foo):
            self.foo = foo

    @mock.patch.object(session_local_storage, '_patch')
    def setUp(self, mock_patch):
        super(TestLocal, self).setUp()
        self.fl = self.FakeLocal('bar')

    @mock.patch.object(session_local_storage, '_patch')
    def test_getattribute(self, mock_patch):
        self.assertEqual('bar', self.fl.foo)
        mock_patch.assert_called_with(self.fl)

    @mock.patch.object(session_local_storage, '_patch')
    def test_setattribute(self, mock_patch):
        self.fl.foo = 'baz'
        mock_patch.assert_called_with(self.fl)

    @mock.patch.object(session_local_storage, '_patch')
    def test_delattribute(self, mock_patch):
        del self.fl.foo
        mock_patch.assert_called_with(self.fl)


class TestSessionLocalDict(test_case.DslTestCase):
    def setUp(self):
        super(TestSessionLocalDict, self).setUp()
        self.sld = session_local_storage.SessionLocalDict(foo='bar')

    @mock.patch.object(helpers, 'get_execution_session', return_value=None)
    def test_data_no_session(self, mock_ges):
        self.assertEqual({'foo': 'bar'}, self.sld.data)
        self.sld.data = {'foo': 'baz'}

        mock_ges.assert_called_with()
        self.assertEqual({'foo': 'baz'}, self.sld.data)

    @mock.patch.object(helpers, 'get_execution_session',
                       return_value=mock.sentinel.session)
    def test_data(self, mock_ges):
        self.sld.data = mock.sentinel.data

        mock_ges.assert_called_with()
        self.assertEqual(mock.sentinel.data, self.sld.data)


class TestExecutionSessionMemoize(test_case.DslTestCase):
    @mock.patch.object(helpers, 'get_memoize_func',
                       return_value=mock.sentinel.mem_func)
    def test_execution_session_memoize(self, mock_gef):
        f = mock.MagicMock()
        f.return_value = 'im a function'
        new_f = session_local_storage.execution_session_memoize(f)
        self.assertEqual(f, mock_gef.call_args[0][0])
        self.assertIsInstance(mock_gef.call_args[0][1],
                              session_local_storage.SessionLocalDict)
        self.assertEqual(mock.sentinel.mem_func, new_f)



#    Copyright (c) 2015 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import inspect
import os.path

import eventlet
from oslo_config import cfg
from yaql.language import expressions as yaql_expressions
from yaql.language import specs
from yaql.language import utils
from yaql.language import yaqltypes
from yaql import yaql_interface

from murano.dsl import constants
from murano.dsl import dsl_types
from murano.dsl import helpers


CONF = cfg.CONF
NO_VALUE = utils.create_marker('NO_VALUE')


def name(dsl_name):
    def wrapper(cls):
        cls.__murano_name = dsl_name
        return cls
    return wrapper


class MuranoObjectParameter(yaqltypes.PythonType):
    def __init__(self, murano_class=None, nullable=False, version_spec=None,
                 decorate=True):
        self.murano_class = murano_class
        self.version_spec = version_spec
        self.decorate = decorate
        super(MuranoObjectParameter, self).__init__(
            (dsl_types.MuranoObject, MuranoObjectInterface), nullable)

    def check(self, value, context, *args, **kwargs):
        if not super(MuranoObjectParameter, self).check(
                value, context, *args, **kwargs):
            return False
        if value is None or isinstance(value, yaql_expressions.Expression):
            return True
        if isinstance(value, MuranoObjectInterface):
            value = value.object
        if not isinstance(value, dsl_types.MuranoObject):
            return False
        if self.murano_class:
            murano_class = self.murano_class
            if isinstance(murano_class, str):
                return helpers.is_instance_of(
                    value, murano_class,
                    self.version_spec or helpers.get_type(context))
            else:
                return murano_class.is_compatible(value)
        else:
            return True

    def convert(self, value, sender, context, function_spec, engine,
                *args, **kwargs):
        result = super(MuranoObjectParameter, self).convert(
            value, sender, context, function_spec, engine, *args, **kwargs)
        if self.decorate:
            return MuranoObjectInterface.create(result)
        else:
            if isinstance(result, dsl_types.MuranoObject):
                return result
            return None if result is None else result.object


class ThisParameter(yaqltypes.HiddenParameterType, yaqltypes.SmartType):
    def __init__(self):
        super(ThisParameter, self).__init__(False)

    def convert(self, value, sender, context, function_spec, engine,
                *args, **kwargs):
        return get_this(context)


class InterfacesParameter(yaqltypes.HiddenParameterType,
                          yaqltypes.SmartType):
    def __init__(self):
        super(InterfacesParameter, self).__init__(False)

    def convert(self, value, sender, context, function_spec, engine,
                *args, **kwargs):
        this = helpers.get_this(context)
        return Interfaces(this)


class MuranoTypeParameter(yaqltypes.PythonType):
    def __init__(self, base_type=None, nullable=False, context=None,
                 resolve_strings=True, lazy=False):
        self._context = context
        self._base_type = base_type
        self._resolve_strings = resolve_strings
        self._lazy = lazy
        super(MuranoTypeParameter, self).__init__(
            (dsl_types.MuranoTypeReference, str), nullable)

    def check(self, value, context, *args, **kwargs):
        if not super(MuranoTypeParameter, self).check(
                value, context, *args, **kwargs):
            return False
        if isinstance(value, str):
            if not self._resolve_strings:
                return False
            value = helpers.get_class(value, context).get_reference()
        if isinstance(value, dsl_types.MuranoTypeReference):
            if not self._base_type:
                return True
            return self._base_type.is_compatible(value)
        return True

    def convert(self, value, sender, context, function_spec, engine,
                *args, **kwargs):

        def implementation(ctx=None):
            value2 = value
            if ctx is None:
                ctx = self._context or context
            if isinstance(value2, yaql_expressions.Expression):
                value2 = value2(utils.NO_VALUE, ctx, engine)
            value2 = super(MuranoTypeParameter, self).convert(
                value2, sender, ctx, function_spec, engine)
            if isinstance(value2, str):
                value2 = helpers.get_class(value2, ctx).get_reference()
            if self._base_type and not self._base_type.is_compatible(value):
                raise ValueError('Value must be subtype of {0}'.format(
                    self._base_type.name
                ))
            return value2

        if self._lazy:
            return implementation
        return implementation()


class MuranoObjectInterface(dsl_types.MuranoObjectInterface):
    class DataInterface(object):
        def __init__(self, object_interface):
            object.__setattr__(self, '__object_interface', object_interface)

        def __getattr__(self, item):
            oi = getattr(self, '__object_interface')
            return oi[item]

        def __setattr__(self, key, value):
            oi = getattr(self, '__object_interface')
            oi[key] = value

    class CallInterface(object):
        def __init__(self, object_interface):
            self.__object_interface = object_interface

        def __getattr__(self, item):
            def func(*args, **kwargs):
                self._insert_instruction()
                with helpers.with_object_store(
                        self.__object_interface._object_store):
                    context = helpers.get_context()
                    obj = self.__object_interface.object
                    return to_mutable(obj.type.invoke(
                        item, obj, args, kwargs,
                        context), helpers.get_yaql_engine(context))
            return func

        @staticmethod
        def _insert_instruction():
            context = helpers.get_context()
            if context:
                frame = inspect.stack()[2]
                location = dsl_types.ExpressionFilePosition(
                    os.path.abspath(frame[1]), frame[2],
                    -1, frame[2], -1)
                context[constants.CTX_CURRENT_INSTRUCTION] = NativeInstruction(
                    frame[4][0].strip(), location)

    def __init__(self, mpl_object):
        self._object = mpl_object
        self._object_store = helpers.get_object_store()

    @staticmethod
    def create(mpl_object):
        if mpl_object is None or isinstance(mpl_object, MuranoObjectInterface):
            return mpl_object
        return MuranoObjectInterface(mpl_object)

    @property
    def object(self):
        return self._object

    @property
    def id(self):
        return self.object.object_id

    @property
    def owner(self):
        owner = self.object.owner
        return MuranoObjectInterface.create(owner)

    def find_owner(self, type, optional=False):
        if isinstance(type, str):
            type = helpers.get_class(type)
        elif isinstance(type, dsl_types.MuranoTypeReference):
            type = type.type
        owner = helpers.find_object_owner(
            self.object, lambda t: type.is_compatible(t))
        if owner:
            return MuranoObjectInterface(owner)
        if not optional:
            raise ValueError('Object is not owned by any instance of type '
                             '{0}'.format(type.name))
        return None

    @property
    def type(self):
        return self.object.type

    @property
    def package(self):
        return self.type.package

    @property
    def properties(self):
        return MuranoObjectInterface.DataInterface(self)

    @property
    def name(self):
        return self.object.name

    @property
    def extension(self):
        return self.object.extension

    def cast(self, murano_class, version_spec=None):
        return MuranoObjectInterface.create(
            helpers.cast(
                self.object, murano_class,
                version_spec or helpers.get_type()))

    def is_instance_of(self, murano_class, version_spec=None):
        return helpers.is_instance_of(
            self.object, murano_class,
            version_spec or helpers.get_type())

    def ancestors(self):
        return self.type.ancestors()

    def __getitem__(self, item):
        context = helpers.get_context()
        return to_mutable(
            self.object.get_property(item, context),
            helpers.get_yaql_engine(context))

    def __setitem__(self, key, value):
        context = helpers.get_context()
        value = helpers.evaluate(value, context)
        self.object.set_property(key, value, context)

    def __call__(self):
        return MuranoObjectInterface.CallInterface(self)

    def __repr__(self):
        return '<{0}>'.format(repr(self.object))

    def __eq__(self, other):
        if isinstance(other, MuranoObjectInterface):
            return self.object == other.object
        else:
            return False

    def __ne__(self, other):
        return not (self == other)

    def __hash__(self):
        return hash(self.object)


class Interfaces(object):
    def __init__(self, mpl_object):
        self.__object = mpl_object

    def yaql(self, receiver=utils.NO_VALUE):
        return yaql_interface.YaqlInterface(
            helpers.get_context(), helpers.get_yaql_engine(), receiver)

    def this(self):
        return self.methods(self.__object)

    def methods(self, mpl_object):
        return MuranoObjectInterface.create(mpl_object)

    @property
    def execution_session(self):
        return helpers.get_execution_session()

    @property
    def caller(self):
        caller_context = helpers.get_caller_context()
        if caller_context is None:
            return None
        return get_this(caller_context)

    @property
    def attributes(self):
        executor = helpers.get_executor()
        return executor.attribute_store

    @property
    def class_config(self):
        return self.__object.type.package.get_class_config(
            self.__object.type.name)

    @property
    def package_loader(self):
        return helpers.get_package_loader()


class NativeInstruction(object):
    def __init__(self, instruction, location):
        self.instruction = instruction
        self.source_file_position = location

    def __str__(self):
        return self.instruction


def to_mutable(obj, yaql_engine=None):
    if yaql_engine is None:
        yaql_engine = helpers.get_yaql_engine()

    def converter(value, limit_func, engine, rec):
        if isinstance(value, dsl_types.MuranoObject):
            return MuranoObjectInterface.create(value)
        else:
            return utils.convert_output_data(value, limit_func, engine, rec)

    def limiter(it):
        return utils.limit_iterable(it, CONF.murano.dsl_iterators_limit)

    return converter(obj, limiter, yaql_engine, converter)


def meta(type_name, value):
    def wrapper(func):
        fd = specs.get_function_definition(func)
        mpl_meta = fd.meta.get(constants.META_MPL_META, [])
        mpl_meta.append({type_name: value})
        specs.meta(type_name, mpl_meta)(func)
    return wrapper


def get_this(context=None):
    this = helpers.get_this(context)
    return MuranoObjectInterface.create(this)


def get_execution_session():
    return helpers.get_execution_session()


def spawn(func, *args, **kwargs):
    context = helpers.get_context()
    object_store = helpers.get_object_store()

    def wrapper():
        with helpers.with_object_store(object_store):
            with helpers.contextual(context):
                return func(*args, **kwargs)

    return eventlet.spawn(wrapper)


def new(properties, owner=None, type=None):
    context = helpers.get_context()
    return helpers.get_object_store().load(
        properties, owner, type or get_this(context).type, context=context)



# Copyright 2017 AT&T Corporation.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_policy import policy

from murano.common.policies import base

template_policies = [
    policy.DocumentedRuleDefault(
        name='list_env_templates',
        check_str=base.RULE_DEFAULT,
        description='List environment templates in a project.',
        operations=[{'path': '/v1/templates',
                     'method': 'GET'}]),
    policy.DocumentedRuleDefault(
        name='create_env_template',
        check_str=base.RULE_DEFAULT,
        description='Create an environment template.',
        operations=[{'path': '/v1/templates',
                     'method': 'POST'}]),
    policy.DocumentedRuleDefault(
        name='show_env_template',
        check_str=base.RULE_DEFAULT,
        description='Show environment template details.',
        operations=[{'path': '/v1/templates/{env_template_id}',
                     'method': 'GET'}]),
    policy.DocumentedRuleDefault(
        name='update_env_template',
        check_str=base.RULE_DEFAULT,
        description='Update an environment template.',
        operations=[{'path': '/v1/templates/{env_template_id}',
                     'method': 'PUT'}]),
    policy.DocumentedRuleDefault(
        name='delete_env_template',
        check_str=base.RULE_DEFAULT,
        description='Delete an environment template.',
        operations=[{'path': '/v1/templates/{env_template_id}',
                     'method': 'DELETE'}]),
    policy.DocumentedRuleDefault(
        name='clone_env_template',
        check_str=base.RULE_DEFAULT,
        description='Clone an environment template.',
        operations=[{'path': '/v1/templates/{env_template_id}/clone',
                     'method': 'POST'}])
]


def list_rules():
    return template_policies



#    Copyright (c) 2014 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import contextlib
import os
import shutil
import sys
import tempfile
import zipfile

import yaml

from murano.common.helpers import path
from murano.common.plugins import package_types_loader
from murano.common import utils
import murano.packages.exceptions as e
import murano.packages.hot_package
import murano.packages.mpl_package


PLUGIN_LOADER = None


def get_plugin_loader():
    global PLUGIN_LOADER

    if PLUGIN_LOADER is None:
        PLUGIN_LOADER = package_types_loader.PluginLoader()
        for runtime_version in ('1.0', '1.1', '1.2', '1.3', '1.4'):
            format_string = 'MuranoPL/' + runtime_version
            PLUGIN_LOADER.register_format(
                format_string, murano.packages.mpl_package.MuranoPlPackage)
        PLUGIN_LOADER.register_format(
            'Heat.HOT/1.0', murano.packages.hot_package.HotPackage)
    return PLUGIN_LOADER


@contextlib.contextmanager
def load_from_file(archive_path, target_dir=None, drop_dir=False):
    if not os.path.isfile(archive_path):
        raise e.PackageLoadError('Unable to find package file')
    created = False
    if not target_dir:
        target_dir = tempfile.mkdtemp()
        created = True
    elif not os.path.exists(target_dir):
        os.makedirs(target_dir)
        created = True
    else:
        if os.listdir(target_dir):
            raise e.PackageLoadError('Target directory is not empty')

    try:
        if not zipfile.is_zipfile(archive_path):
            raise e.PackageFormatError("Uploaded file {0} is not a "
                                       "zip archive".format(archive_path))
        package = zipfile.ZipFile(archive_path)
        package.extractall(path=target_dir)
        yield load_from_dir(target_dir)
    except ValueError as err:
        raise e.PackageLoadError("Couldn't load package from file: "
                                 "{0}".format(err))
    finally:
        if drop_dir:
            if created:
                shutil.rmtree(target_dir)
            else:
                for f in os.listdir(target_dir):
                    os.unlink(path.secure_join(target_dir, f))


def load_from_dir(source_directory, filename='manifest.yaml'):
    if not os.path.isdir(source_directory) or not os.path.exists(
            source_directory):
        raise e.PackageLoadError('Invalid package directory')
    full_path = path.secure_join(source_directory, filename)
    if not os.path.isfile(full_path):
        raise e.PackageLoadError('Unable to find package manifest')

    try:
        with open(full_path) as stream:
            content = yaml.safe_load(stream)
    except Exception as ex:
        trace = sys.exc_info()[2]
        utils.reraise(
            e.PackageLoadError,
            e.PackageLoadError("Unable to load due to '{0}'".format(ex)),
            trace)
    else:
        format_spec = str(content.get('Format') or 'MuranoPL/1.0')
        if format_spec[0].isdigit():
            format_spec = 'MuranoPL/' + format_spec
        plugin_loader = get_plugin_loader()
        handler = plugin_loader.get_package_handler(format_spec)
        if handler is None:
            raise e.PackageFormatError(
                'Unsupported format {0}'.format(format_spec))
        return handler(source_directory, content)



# Copyright (c) 2016 AT&T Corp
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from unittest import mock

from keystoneauth1 import loading as ka_loading

from murano.common import auth_utils
from murano.tests.unit import base

from oslo_config import cfg


class TestAuthUtils(base.MuranoTestCase):

    def setUp(self):
        super(TestAuthUtils, self).setUp()
        # Register the Password auth plugin options,
        # so we can use CONF.set_override
        password_option = ka_loading.get_auth_plugin_conf_options('password')
        cfg.CONF.register_opts(password_option,
                               group=auth_utils.CFG_MURANO_AUTH_GROUP)
        self.addCleanup(mock.patch.stopall)

    def _init_mock_cfg(self):
        mock_auth_obj = mock.patch.object(auth_utils, 'ka_loading',
                                          spec_set=ka_loading).start()
        mock_auth_obj.load_auth_from_conf_options.return_value = \
            mock.sentinel.auth
        mock_auth_obj.load_session_from_conf_options.\
            return_value = mock.sentinel.session
        cfg.CONF.set_override('auth_type',
                              'password',
                              auth_utils.CFG_MURANO_AUTH_GROUP)
        cfg.CONF.set_override('www_authenticate_uri',
                              'foo_www_authenticate_uri',
                              auth_utils.CFG_MURANO_AUTH_GROUP)
        cfg.CONF.set_override('auth_url',
                              'foo_auth_url',
                              auth_utils.CFG_MURANO_AUTH_GROUP)
        cfg.CONF.set_override('username',
                              'fakeuser',
                              auth_utils.CFG_MURANO_AUTH_GROUP)
        cfg.CONF.set_override('password',
                              'fakepass',
                              auth_utils.CFG_MURANO_AUTH_GROUP)
        cfg.CONF.set_override('user_domain_name',
                              'Default',
                              auth_utils.CFG_MURANO_AUTH_GROUP)
        cfg.CONF.set_override('project_domain_name',
                              'Default',
                              auth_utils.CFG_MURANO_AUTH_GROUP)
        cfg.CONF.set_override('project_name',
                              'fakeproj',
                              auth_utils.CFG_MURANO_AUTH_GROUP)
        return mock_auth_obj

    def test_get_keystone_auth(self):
        mock_identity = self._init_mock_cfg()

        expected_auth = mock.sentinel.auth
        actual_auth = auth_utils._get_keystone_auth()

        self.assertEqual(expected_auth, actual_auth)
        mock_identity.load_auth_from_conf_options.assert_called_once_with(
            cfg.CONF, auth_utils.CFG_MURANO_AUTH_GROUP)

    def test_get_keystone_with_trust_id(self):
        mock_ka_loading = self._init_mock_cfg()

        expected_kwargs = {
            'project_name': None,
            'project_domain_name': None,
            'project_id': None,
            'trust_id': mock.sentinel.trust_id
        }
        expected_auth = mock.sentinel.auth
        actual_auth = auth_utils._get_keystone_auth(mock.sentinel.trust_id)

        self.assertEqual(expected_auth, actual_auth)
        mock_ka_loading.load_auth_from_conf_options.assert_called_once_with(
            cfg.CONF,
            auth_utils.CFG_MURANO_AUTH_GROUP,
            **expected_kwargs)

    @mock.patch.object(auth_utils, 'ks_client', autospec=True)
    @mock.patch.object(auth_utils, '_get_session', autospec=True)
    def test_create_keystone_admin_client(self, mock_get_sess, mock_ks_client):
        self._init_mock_cfg()
        mock_get_sess.return_value = mock.sentinel.session
        mock_ks_client.Client.return_value = mock.sentinel.ks_admin_client

        result = auth_utils._create_keystone_admin_client()

        self.assertEqual(result, mock.sentinel.ks_admin_client)
        self.assertTrue(mock_get_sess.called)
        mock_ks_client.Client.assert_called_once_with(
            session=mock.sentinel.session)

    @mock.patch.object(auth_utils, '_get_session', autospec=True)
    @mock.patch.object(auth_utils, '_get_keystone_auth', autospec=True)
    @mock.patch.object(
        auth_utils.helpers, 'get_execution_session', autospec=True)
    def test_get_client_session(self, mock_get_execution_session,
                                mock_get_keystone_auth, mock_get_session):
        mock_exec_session = mock.Mock(trust_id=mock.sentinel.trust_id)
        mock_get_execution_session.return_value = mock_exec_session
        mock_get_keystone_auth.return_value = mock.sentinel.auth
        mock_get_session.return_value = mock.sentinel.session

        session = auth_utils.get_client_session(conf=mock.sentinel.conf)

        self.assertEqual(mock.sentinel.session, session)
        mock_get_execution_session.assert_called_once_with()
        mock_get_keystone_auth.assert_called_once_with(mock.sentinel.trust_id)
        mock_get_session.assert_called_once_with(
            auth=mock.sentinel.auth,
            conf_section=mock.sentinel.conf)

    @mock.patch.object(auth_utils, 'get_token_client_session', autospec=True)
    @mock.patch.object(
        auth_utils.helpers, 'get_execution_session', autospec=True)
    def test_get_client_session_without_trust_id(
            self, mock_get_execution_session, mock_get_token_client_session):
        mock_get_token_client_session.return_value = mock.sentinel.session
        mock_exec_session = mock.Mock(trust_id=None,
                                      token=mock.sentinel.token,
                                      project_id=mock.sentinel.project_id)

        session = auth_utils.get_client_session(
            execution_session=mock_exec_session, conf=mock.sentinel.conf)

        self.assertEqual(mock.sentinel.session, session)
        self.assertFalse(mock_get_execution_session.called)
        mock_get_token_client_session.assert_called_once_with(
            token=mock.sentinel.token, project_id=mock.sentinel.project_id)

    @mock.patch.object(auth_utils, '_get_session', autospec=True)
    @mock.patch.object(auth_utils, 'identity', autospec=True)
    @mock.patch.object(
        auth_utils.helpers, 'get_execution_session', autospec=True)
    def test_get_token_client_session(
            self, mock_get_execution_session, mock_identity,
            mock_get_session):
        cfg.CONF.set_override('www_authenticate_uri',
                              'foo_www_authenticate_uri/v2.0',
                              auth_utils.CFG_MURANO_AUTH_GROUP)

        mock_get_execution_session.return_value = \
            mock.Mock(token=mock.sentinel.token,
                      project_id=mock.sentinel.project_id)
        mock_identity.Token.return_value = mock.sentinel.auth
        mock_get_session.return_value = mock.sentinel.session

        session = auth_utils.get_token_client_session()
        self.assertEqual(mock.sentinel.session, session)

        mock_get_execution_session.assert_called_once_with()
        mock_identity.Token.assert_called_once_with(
            'foo_www_authenticate_uri/v3', token=mock.sentinel.token,
            project_id=mock.sentinel.project_id)
        mock_get_session.assert_called_once_with(
            auth=mock.sentinel.auth,
            conf_section=None)

    @mock.patch.object(auth_utils, 'get_token_client_session', autospec=True)
    @mock.patch.object(auth_utils, 'ks_client', autospec=True)
    def test_create_keystone_client(self, mock_ks_client,
                                    mock_get_token_client_session):
        mock_ks_client.Client.return_value = mock.sentinel.ks_client
        mock_session = mock.Mock(
            token=mock.sentinel.token, project_id=mock.sentinel.project_id,
            conf=mock.sentinel.conf)
        mock_get_token_client_session.return_value = mock_session

        ks_client = auth_utils.create_keystone_client(
            mock.sentinel.token, mock.sentinel.project_id, mock.sentinel.conf)

        self.assertEqual(mock.sentinel.ks_client, ks_client)
        mock_ks_client.Client.assert_called_once_with(session=mock_session)

    @mock.patch.object(auth_utils, 'create_keystone_client', autospec=True)
    @mock.patch.object(
        auth_utils, '_create_keystone_admin_client', autospec=True)
    def test_create_trust(self, mock_create_ks_admin_client,
                          mock_create_ks_client):
        mock_auth_ref = mock.Mock(user_id=mock.sentinel.trustor_user,
                                  project_id=mock.sentinel.project_id,
                                  role_names=mock.sentinel.role_names)
        mock_admin_session = mock.Mock(**{
            'auth.get_user_id.return_value': mock.sentinel.trustee_user
        })
        mock_user_session = mock.Mock(**{
            'auth.get_access.return_value': mock_auth_ref
        })
        mock_trust = mock.Mock(id=mock.sentinel.trust_id)
        mock_admin_client = mock.Mock(session=mock_admin_session)
        mock_user_client = mock.Mock(
            session=mock_user_session,
            **{'trusts.create.return_value': mock_trust})

        mock_create_ks_admin_client.return_value = mock_admin_client
        mock_create_ks_client.return_value = mock_user_client

        trust_id = auth_utils.create_trust(
            trustee_token=mock.sentinel.trustee_token,
            trustee_project_id=mock.sentinel.trustee_project_id)

        self.assertEqual(mock.sentinel.trust_id, trust_id)
        mock_create_ks_admin_client.assert_called_once_with()
        mock_create_ks_client.assert_called_once_with(
            token=mock.sentinel.trustee_token,
            project_id=mock.sentinel.trustee_project_id)
        mock_admin_client.session.auth.get_user_id.assert_called_once_with(
            mock_admin_session)
        mock_user_client.session.auth.get_access.assert_called_once_with(
            mock_user_session)
        mock_user_client.trusts.create.assert_called_once_with(
            trustor_user=mock.sentinel.trustor_user,
            trustee_user=mock.sentinel.trustee_user,
            impersonation=True,
            role_names=mock.sentinel.role_names,
            project=mock.sentinel.project_id)

    @mock.patch.object(auth_utils, 'create_keystone_client', autospec=True)
    def test_delete_trust(self, mock_ks_client):
        mock_auth_ref = mock.Mock(trust_id=mock.sentinel.trust_id,
                                  token=mock.sentinel.token,
                                  project_id=mock.sentinel.project_id)
        mock_user_session = mock.Mock(**{
            'auth.get_access.return_value': mock_auth_ref
        })
        mock_user_client = mock.Mock(
            session=mock_user_session)

        mock_ks_client.return_value = mock_user_client

        auth_utils.delete_trust(mock_auth_ref)

        mock_user_client.trusts.delete.assert_called_once_with(
            mock_auth_ref.trust_id)

    def test_get_config_option(self):
        cfg.CONF.set_override('url', 'foourl', 'murano')
        self.assertEqual('foourl', auth_utils._get_config_option(
            'murano', 'url'))

    def test_get_config_option_return_default(self):
        self.assertIsNone(auth_utils._get_config_option(None, 'url'))

    def test_get_session(self):
        mock_ka_loading = self._init_mock_cfg()

        session = auth_utils._get_session(mock.sentinel.auth)

        self.assertEqual(mock.sentinel.session, session)
        mock_ka_loading.load_session_from_conf_options.\
            assert_called_once_with(auth=mock.sentinel.auth,
                                    conf=cfg.CONF,
                                    group=auth_utils.CFG_MURANO_AUTH_GROUP)

    def test_get_session_client_parameters(self):

        cfg.CONF.set_override('url', 'foourl', 'murano')

        expected_result = {
            'session': mock.sentinel.session,
            'endpoint_override': 'foourl'
        }

        result = auth_utils.get_session_client_parameters(
            conf='murano',
            service_type=mock.sentinel.service_type,
            service_name=mock.sentinel.service_name,
            session=mock.sentinel.session)

        for key, val in expected_result.items():
            self.assertEqual(val, result[key])

    def test_get_session_client_parameters_without_url(self):
        cfg.CONF.set_override('home_region', 'fooregion')

        expected_result = {
            'session': mock.sentinel.session,
            'service_type': mock.sentinel.service_type,
            'service_name': mock.sentinel.service_name,
            'interface': mock.sentinel.endpoint_type,
            'region_name': 'fooregion'
        }

        result = auth_utils.get_session_client_parameters(
            service_type=mock.sentinel.service_type,
            service_name=mock.sentinel.service_name,
            interface=mock.sentinel.endpoint_type,
            session=mock.sentinel.session)

        for key, val in expected_result.items():
            self.assertEqual(val, result[key])

    @mock.patch.object(
        auth_utils, '_create_keystone_admin_client', autospec=True)
    def test_get_user(self, mock_create_ks_admin_client):
        mock_client = mock.Mock(
            **{'users.get.return_value.to_dict.return_value':
                mock.sentinel.user})
        mock_create_ks_admin_client.return_value = mock_client

        user = auth_utils.get_user(mock.sentinel.uid)

        self.assertEqual(mock.sentinel.user, user)
        mock_client.users.get.assert_called_once_with(mock.sentinel.uid)
        mock_client.users.get.return_value.to_dict.assert_called_once_with()

    @mock.patch.object(
        auth_utils, '_create_keystone_admin_client', autospec=True)
    def test_get_project(self, mock_create_ks_admin_client):
        mock_client = mock.Mock(
            **{'projects.get.return_value.to_dict.return_value':
                mock.sentinel.project})
        mock_create_ks_admin_client.return_value = mock_client

        project = auth_utils.get_project(mock.sentinel.pid)

        self.assertEqual(mock.sentinel.project, project)
        mock_client.projects.get.assert_called_once_with(mock.sentinel.pid)
        mock_client.projects.get.return_value.to_dict.assert_called_once_with()



# Copyright (c) 2014 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import murano.common.exceptions as e


class PackageException(e.Error):
    pass


class PackageClassLoadError(PackageException):
    def __init__(self, class_name, message=None):
        msg = 'Unable to load class "{0}" from package'.format(class_name)
        if message:
            msg += ": " + message
        super(PackageClassLoadError, self).__init__(msg)


class PackageUILoadError(PackageException):
    def __init__(self, message=None):
        msg = 'Unable to load ui definition from package'
        if message:
            msg += ": " + message
        super(PackageUILoadError, self).__init__(msg)


class PackageLoadError(PackageException):
    pass


class PackageFormatError(PackageLoadError):
    def __init__(self, message=None):
        msg = 'Incorrect package format'
        if message:
            msg += ': ' + message
        super(PackageFormatError, self).__init__(msg)



# Copyright (c) 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

from keystoneauth1 import identity
from keystoneauth1 import loading as ka_loading
from keystoneclient.v3 import client as ks_client
from oslo_config import cfg
from oslo_log import log as logging
from oslo_log import versionutils

from murano.dsl import helpers


CFG_KEYSTONE_GROUP = 'keystone_authtoken'
CFG_MURANO_AUTH_GROUP = 'murano_auth'
LOG = logging.getLogger(__name__)

cfg.CONF.import_group(CFG_KEYSTONE_GROUP, 'keystonemiddleware.auth_token')


def _get_keystone_auth(trust_id=None):
    kwargs = {}
    if trust_id:
        # Remove project_name and project_id, since we need a trust scoped
        # auth object
        kwargs['project_name'] = None
        kwargs['project_domain_name'] = None
        kwargs['project_id'] = None
        kwargs['trust_id'] = trust_id
    auth = ka_loading.load_auth_from_conf_options(
        cfg.CONF,
        CFG_MURANO_AUTH_GROUP,
        **kwargs)
    return auth


def _create_keystone_admin_client():
    auth = _get_keystone_auth()
    session = _get_session(auth=auth)
    return ks_client.Client(session=session)


def get_client_session(execution_session=None, conf=None):
    if not execution_session:
        execution_session = helpers.get_execution_session()
    trust_id = execution_session.trust_id
    if trust_id is None:
        return get_token_client_session(
            token=execution_session.token,
            project_id=execution_session.project_id)
    auth = _get_keystone_auth(trust_id)
    session = _get_session(auth=auth, conf_section=conf)
    return session


def get_token_client_session(token=None, project_id=None, conf=None):
    www_authenticate_uri = \
        cfg.CONF[CFG_MURANO_AUTH_GROUP].www_authenticate_uri
    if not www_authenticate_uri:
        versionutils.report_deprecated_feature(
            LOG, 'Please configure www_authenticate_uri in ' +
            CFG_MURANO_AUTH_GROUP + 'group')
        www_authenticate_uri = \
            cfg.CONF[CFG_KEYSTONE_GROUP].www_authenticate_uri
    if not (www_authenticate_uri.endswith('v2.0') or
            www_authenticate_uri.endswith('v3')):
        auth_url = os.path.join(www_authenticate_uri, 'v3')
    elif www_authenticate_uri.endswith('v2.0'):
        auth_url = www_authenticate_uri.replace('v2.0', 'v3')
    else:
        auth_url = www_authenticate_uri

    if token is None or project_id is None:
        execution_session = helpers.get_execution_session()
        token = execution_session.token
        project_id = execution_session.project_id
    token_auth = identity.Token(
        auth_url,
        token=token,
        project_id=project_id)
    session = _get_session(auth=token_auth, conf_section=conf)
    return session


def create_keystone_client(token=None, project_id=None, conf=None):
    return ks_client.Client(session=get_token_client_session(
        token=token, project_id=project_id, conf=conf))


def create_trust(trustee_token=None, trustee_project_id=None):
    admin_client = _create_keystone_admin_client()
    user_client = create_keystone_client(
        token=trustee_token, project_id=trustee_project_id)
    trustee_user = admin_client.session.auth.get_user_id(admin_client.session)
    auth_ref = user_client.session.auth.get_access(user_client.session)
    trustor_user = auth_ref.user_id
    project = auth_ref.project_id
    roles = auth_ref.role_names
    trust = user_client.trusts.create(
        trustor_user=trustor_user,
        trustee_user=trustee_user,
        impersonation=True,
        role_names=roles,
        project=project)
    return trust.id


def delete_trust(session):
    user_client = create_keystone_client(
        token=session.token, project_id=session.project_id)
    user_client.trusts.delete(session.trust_id)


def _get_config_option(conf_section, option_name, default=None):
    if conf_section and hasattr(cfg.CONF[conf_section], option_name):
        return getattr(cfg.CONF[conf_section], option_name)
    return default


def _get_session(auth, conf_section=None):
    # Fallback to murano_auth section for TLS parameters
    # if no other conf_section supplied
    if not conf_section:
        conf_section = CFG_MURANO_AUTH_GROUP
    session = ka_loading.load_session_from_conf_options(
        auth=auth,
        conf=cfg.CONF,
        group=conf_section)
    return session


def get_session_client_parameters(service_type=None,
                                  region='',
                                  interface=None,
                                  service_name=None,
                                  conf=None,
                                  session=None,
                                  execution_session=None):
    if region == '':
        region = cfg.CONF.home_region
    result = {
        'session': session or get_client_session(
            execution_session=execution_session, conf=conf)
    }

    url = _get_config_option(conf, 'url')
    if url:
        result['endpoint_override'] = url
    else:
        if not interface:
            interface = _get_config_option(conf, 'endpoint_type')
        result.update({
            'service_type': service_type,
            'service_name': service_name,
            'interface': interface,
            'region_name': region
        })
    return result


def get_user(uid):
    client = _create_keystone_admin_client()
    return client.users.get(uid).to_dict()


def get_project(pid):
    client = _create_keystone_admin_client()
    return client.projects.get(pid).to_dict()



# Copyright 2016 AT&T Corp
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import yaml

from murano.packages import exceptions
import murano.packages.hot_package
import murano.packages.load_utils as load_utils
import murano.tests.unit.base as test_base


class TestHotPackage(test_base.MuranoTestCase):

    def _get_hot_package(self, source_directory):
        manifest = {
            'FullName': 'FullTestName',
            'Version': '1.0.0',
            'Type': 'Application',
            'Name': 'TestName',
            'Description': 'TestDescription',
            'Author': 'TestAuthor',
            'Supplier': 'TestSupplier',
            'Logo:': 'TestLogo',
            'Tags': ['Tag1', 'Tag2']
        }
        return murano.packages.hot_package.HotPackage(
            None, None, source_directory=source_directory,
            manifest=manifest
        )

    @classmethod
    def setUpClass(cls):
        super(TestHotPackage, cls).setUpClass()
        this_dir = os.path.dirname(os.path.realpath(__file__))
        cls.test_dirs = [
            os.path.join(this_dir, 'test.hot.1'),
            os.path.join(this_dir, 'test.hot.2'),
            os.path.join(this_dir, 'test.hot.3')
        ]

        manifest_path = os.path.join(cls.test_dirs[0], 'template.yaml')
        cls.manifest = {}
        with open(manifest_path) as manifest_file:
            for key, value in yaml.safe_load(manifest_file).items():
                cls.manifest[key] = value

        properties_manifest_path = os.path.join(cls.test_dirs[0],
                                                'properties_manifest.yaml')
        cls.properties_manifest = {}
        with open(properties_manifest_path) as manifest_file:
            for key, value in yaml.safe_load(manifest_file).items():
                cls.properties_manifest[key] = value

    def test_heat_files_generated(self):
        package_dir = os.path.abspath(
            os.path.join(__file__,
                         '../../test_packages/test.hot.v1.app_with_files')
        )
        load_utils.load_from_dir(package_dir)

        files = murano.packages.hot_package.HotPackage._translate_files(
            package_dir)
        expected_result = {
            "testHeatFile",
            "middle_file/testHeatFile",
            "middle_file/inner_file/testHeatFile",
            "middle_file/inner_file2/testHeatFile"
        }
        msg = "hot files were not generated correctly"
        self.assertSetEqual(expected_result, set(files), msg)

    def test_heat_files_generated_empty(self):
        package_dir = os.path.abspath(
            os.path.join(__file__,
                         '../../test_packages/test.hot.v1.app')
        )
        load_utils.load_from_dir(package_dir)

        files = murano.packages.hot_package.HotPackage \
            ._translate_files(package_dir)
        msg = "heat files were not generated correctly. Expected empty list"
        self.assertEqual([], files, msg)

    def test_build_properties(self):
        result = murano.packages.hot_package.HotPackage._build_properties(
            self.properties_manifest,
            validate_hot_parameters=True)

        self.assertIn('templateParameters', result)
        params = result['templateParameters']
        self.assertEqual(6, len(params['Contract']))

        param1 = params['Contract']['param1']
        param2 = params['Contract']['param2']
        param3 = params['Contract']['param3']
        param4 = params['Contract']['param4']
        param5 = params['Contract']['param5']
        param6 = params['Contract']['param6']

        self.assertEqual("$.bool().check($ in list(True, False))", param1.expr)
        self.assertEqual("$.string().check($ in list('bar'))."
                         "check(len($) <= 50).check(len($) >= 0)."
                         "check(matches($, '[A-Za-z0-9]'))", param2.expr)
        self.assertEqual("$.int().check($ in list(0, 1, 2, 3, 4))"
                         ".check(len($) >= 0 and len($) <= 5)."
                         "check($ >= 0 and $ <= 4)", param3.expr)
        self.assertEqual("$.int().check($ >= -1000).check($ <= "
                         "1000)", param4.expr)
        self.assertEqual("$.string()", param5.expr)
        self.assertEqual("$.string()", param6.expr)

        result = murano.packages.hot_package.HotPackage._build_properties(
            self.properties_manifest,
            validate_hot_parameters=False)
        expected_result = {
            'Contract': {},
            'Default': {},
            'Usage': 'In'
        }
        self.assertEqual(expected_result, result['templateParameters'])

    def test_translate_param_to_contract_with_inappropriate_value(self):
        self.assertRaisesRegex(
            ValueError,
            'Unsupported parameter type',
            murano.packages.hot_package.HotPackage.
            _translate_param_to_contract,
            {'type': 'Inappropriate value'}
        )

    def test_get_class_name(self):
        hot_package = self._get_hot_package(self.test_dirs[0])
        translated_class, _ = hot_package.get_class(hot_package.full_name)
        self.assertIsNotNone(translated_class)
        self.assertEqual(translated_class, hot_package._translated_class)

    def test_get_class_name_with_invalid_template_name(self):
        hot_package = self._get_hot_package(self.test_dirs[0])
        self.assertRaisesRegex(
            exceptions.PackageClassLoadError,
            'Class not defined in this package',
            hot_package.get_class,
            None)

    def test_get_class_name_with_invalid_template_format(self):
        hot_package = self._get_hot_package(self.test_dirs[1])
        self.assertRaisesRegex(
            exceptions.PackageFormatError,
            'Not a HOT template',
            hot_package.get_class,
            hot_package.full_name)

    def test_translate_ui(self):
        hot_package = self._get_hot_package(self.test_dirs[0])
        yaml = hot_package._translate_ui()
        self.assertIsNotNone(yaml)
        expected_application = '''
            "Application":
              "?":
                "classVersion": "1.0.0"
                "package": "FullTestName"
                "type": "FullTestName"
              "name": !yaql "$.group0.name"
              "templateParameters":
                "bar": !yaql "$.group1.bar"
                "baz": !yaql "$.group1.baz"
                "foo": !yaql "$.group1.foo"
        '''
        self.assertIn(expected_application.replace(' ', '').replace('\n', ''),
                      yaml.replace(' ', '').replace('\n', ''))

    def test_translate_ui_with_nonexistent_template(self):
        hot_package = self._get_hot_package(self.test_dirs[2])
        self.assertRaisesRegex(
            exceptions.PackageClassLoadError,
            'File with class definition not found',
            hot_package._translate_ui)

    def test_translate_class_with_nonexistent_template(self):
        hot_package = self._get_hot_package(self.test_dirs[2])
        self.assertRaisesRegex(
            exceptions.PackageClassLoadError,
            'File with class definition not found',
            hot_package._translate_class)



# Copyright (c) 2016 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import eventlet


from murano.tests.unit.dsl.foundation import object_model as om
from murano.tests.unit.dsl.foundation import test_case


class TestConcurrency(test_case.DslTestCase):
    def setUp(self):
        super(TestConcurrency, self).setUp()

        def yield_():
            self.traces.append('yield')
            eventlet.sleep(0)

        self.register_function(yield_, 'yield')
        self._runner = self.new_runner(om.Object('TestConcurrency'))

    def check_isolated_traces(self):
        for i in range(0, len(self.traces), 3):
            before = self.traces[i]
            switch = self.traces[i + 1]
            after = self.traces[i + 2]
            self.assertEqual('yield', switch)
            self.assertEqual(before[0:6], after[0:6])
            self.assertTrue(before.endswith('-before'))
            self.assertTrue(after.endswith('-after'))

    def check_concurrent_traces(self):
        self.assertTrue(self.traces[0].endswith('-before'))
        self.assertEqual('yield', self.traces[1])
        self.assertTrue(self.traces[2].endswith('-before'))
        self.assertEqual('yield', self.traces[3])
        self.assertTrue(self.traces[4].endswith('-before'))
        self.assertEqual('yield', self.traces[5])
        self.assertTrue(self.traces[6].endswith('-after'))
        self.assertTrue(self.traces[7].endswith('-after'))
        self.assertTrue(self.traces[8].endswith('-after'))

    def test_isolated(self):
        self._runner.testCallIsolated()
        self.check_isolated_traces()

    def test_isolated_default(self):
        self._runner.testCallIsolatedWithDefault()
        self.check_isolated_traces()

    def test_concurrent_explicit(self):
        self._runner.testCallConcurrentExplicit()
        self.check_concurrent_traces()

    def test_isolated_explicit(self):
        self._runner.testCallIsolatedExplicit()
        self.check_isolated_traces()

    def test_argbased_primitive_isolated(self):
        self._runner.testCallArgbasedPrimitiveIsolated()
        self.check_isolated_traces()

    def test_argbased_primitive_concurrent(self):
        self._runner.testCallArgbasedPrimitiveConcurrent()
        self.check_concurrent_traces()

    def test_argbased_object_isolated(self):
        self._runner.testCallArgbasedWithObjectIsolated()
        self.check_isolated_traces()

    def test_argbased_object_concurrent(self):
        self._runner.testCallArgbasedWithObjectConcurrent()
        self.check_concurrent_traces()



# Copyright 2016 AT&T Corp
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from datetime import datetime
from unittest import mock

from murano.common import server
from murano.services import states
from murano.tests.unit import base
from murano.tests.unit import utils as test_utils


class ServerTest(base.MuranoTestCase):

    @classmethod
    def setUpClass(cls):
        super(ServerTest, cls).setUpClass()
        cls.result_endpoint = server.ResultEndpoint()
        cls.dummy_context = test_utils.dummy_context()

    @mock.patch('murano.common.server.status_reporter.get_notifier')
    @mock.patch('murano.common.server.LOG')
    @mock.patch('murano.common.server.get_last_deployment')
    @mock.patch('murano.common.server.models')
    @mock.patch('murano.common.server.session')
    def test_process_result(self, mock_db_session, mock_models,
                            mock_last_deployment, mock_log, mock_notifier):
        test_result = {
            'model': {
                'Objects': {
                    'applications': ['app1', 'app2'],
                    'services': ['service1', 'service2']
                }
            },
            'action': {
                'isException': False
            }
        }
        mock_env = mock.MagicMock(id='test_env_id',
                                  tenant_id='test_tenant_id',
                                  description=None,
                                  version=1)
        mock_db_session.get_session().query().get.return_value = mock_env
        mock_db_session.get_session().query().filter_by().count.\
            return_value = 0

        self.result_endpoint.process_result(self.dummy_context, test_result,
                                            'test_env_id')

        self.assertEqual(mock_env.description, test_result['model'])
        self.assertEqual(2, mock_env.version)
        self.assertEqual(test_result['action'],
                         mock_last_deployment().result)
        self.assertEqual('Deployment finished',
                         mock_models.Status().text)
        self.assertEqual('info', mock_models.Status().level)
        mock_last_deployment().statuses.append.assert_called_once_with(
            mock_models.Status())
        mock_db_session.get_session().query().filter_by.assert_any_call(
            **{'environment_id': mock_env.id,
               'state': states.SessionState.DEPLOYING})
        self.assertEqual(
            states.SessionState.DEPLOYED,
            mock_db_session.get_session().query().filter_by().first().state)
        mock_log.info.assert_called_once_with(
            'EnvId: {env_id} TenantId: {tenant_id} Status: '
            'Successful Apps: {services}'
            .format(env_id=mock_env.id,
                    tenant_id=mock_env.tenant_id,
                    services=test_result['model']['Objects']['services']))
        mock_notifier.return_value.report.assert_called_once_with(
            'environment.deploy.end',
            mock_db_session.get_session().query().get(mock_env.id).to_dict())

    @mock.patch('murano.common.server.LOG')
    @mock.patch('murano.common.server.get_last_deployment')
    @mock.patch('murano.common.server.models')
    @mock.patch('murano.common.server.session')
    def test_process_result_with_errors(self, mock_db_session, mock_models,
                                        mock_last_deployment, mock_log):
        test_result = {
            'model': {
                'Objects': {
                    'applications': ['app1', 'app2'],
                    'services': ['service1', 'service2']
                }
            },
            'action': {
                'isException': True
            }
        }
        mock_env = mock.MagicMock(id='test_env_id',
                                  tenant_id='test_tenant_id',
                                  description=None,
                                  version=1)
        mock_db_session.get_session().query().get.return_value = mock_env
        mock_db_session.get_session().query().filter_by().count.\
            return_value = 1

        self.result_endpoint.process_result(self.dummy_context, test_result,
                                            'test_env_id')

        self.assertEqual(mock_env.description, test_result['model'])
        self.assertEqual(test_result['action'],
                         mock_last_deployment().result)
        self.assertEqual('Deployment finished with errors',
                         mock_models.Status().text)
        mock_last_deployment().statuses.append.assert_called_once_with(
            mock_models.Status())
        mock_db_session.get_session().query().filter_by.assert_any_call(
            **{'environment_id': mock_env.id,
               'state': states.SessionState.DEPLOYING})
        self.assertEqual(
            states.SessionState.DEPLOY_FAILURE,
            mock_db_session.get_session().query().filter_by().first().state)
        mock_log.warning.assert_called_once_with(
            'EnvId: {env_id} TenantId: {tenant_id} Status: '
            'Failed Apps: {services}'
            .format(env_id=mock_env.id,
                    tenant_id=mock_env.tenant_id,
                    services=test_result['model']['Objects']['services']))

    @mock.patch('murano.common.server.LOG')
    @mock.patch('murano.common.server.get_last_deployment')
    @mock.patch('murano.common.server.models')
    @mock.patch('murano.common.server.session')
    def test_process_result_with_warnings(self, mock_db_session, mock_models,
                                          mock_last_deployment, mock_log):
        test_result = {
            'model': {
                'Objects': None,
                'ObjectsCopy': ['object1', 'object2']
            },
            'action': {
                'isException': True
            }
        }
        mock_env = mock.MagicMock(id='test_env_id',
                                  tenant_id='test_tenant_id',
                                  description=None,
                                  version=1)
        mock_db_session.get_session().query().get.return_value = mock_env
        # num_errors will be initialized to 0, num_warnings to 1
        mock_db_session.get_session().query().filter_by().count.\
            side_effect = [0, 1]

        self.result_endpoint.process_result(self.dummy_context, test_result,
                                            'test_env_id')

        self.assertEqual(mock_env.description, test_result['model'])
        self.assertEqual(test_result['action'],
                         mock_last_deployment().result)
        self.assertEqual('Deletion finished with warnings',
                         mock_models.Status().text)
        mock_last_deployment().statuses.append.assert_called_once_with(
            mock_models.Status())
        mock_db_session.get_session().query().filter_by.assert_any_call(
            **{'environment_id': mock_env.id,
               'state': states.SessionState.DELETING})
        self.assertEqual(
            states.SessionState.DELETE_FAILURE,
            mock_db_session.get_session().query().filter_by().first().state)
        mock_log.warning.assert_called_once_with(
            'EnvId: {env_id} TenantId: {tenant_id} Status: '
            'Failed Apps: {services}'
            .format(env_id=mock_env.id,
                    tenant_id=mock_env.tenant_id,
                    services=[]))

    @mock.patch('murano.common.server.LOG')
    @mock.patch('murano.common.server.session')
    def test_process_result_with_no_environment(self, mock_db_session,
                                                mock_log):
        test_result = {'model': None}
        mock_db_session.get_session().query().get.return_value = None

        result = self.result_endpoint.process_result(self.dummy_context,
                                                     test_result,
                                                     'test_env_id')
        self.assertIsNone(result)
        mock_log.warning.assert_called_once_with(
            'Environment result could not be handled, '
            'specified environment not found in database')

    @mock.patch('murano.common.server.environments')
    @mock.patch('murano.common.server.session')
    def test_process_result_with_no_objects(self, mock_db_session,
                                            mock_environments):
        test_result = {'model': {'Objects': None, 'ObjectsCopy': None}}

        result = self.result_endpoint.process_result(self.dummy_context,
                                                     test_result,
                                                     'test_env_id')
        self.assertIsNone(result)
        mock_environments.EnvironmentServices.remove.assert_called_once_with(
            'test_env_id')

    @mock.patch('murano.common.server.instances')
    def test_track_instance(self, mock_instances):
        test_payload = {
            'instance': 'test_instance',
            'instance_type': 'test_instance_type',
            'environment': 'test_environment',
            'unit_count': 'test_unit_count',
            'type_name': 'test_type_name',
            'type_title': 'test_type_title'
        }
        server.track_instance(test_payload)
        mock_instances.InstanceStatsServices.track_instance.\
            assert_called_once_with(test_payload['instance'],
                                    test_payload['environment'],
                                    test_payload['instance_type'],
                                    test_payload['type_name'],
                                    test_payload['type_title'],
                                    test_payload['unit_count'])

    @mock.patch('murano.common.server.instances')
    def test_untrack_instance(self, mock_instances):
        test_payload = {
            'instance': 'test_instance',
            'environment': 'test_environment'
        }
        server.untrack_instance(test_payload)
        mock_instances.InstanceStatsServices.destroy_instance.\
            assert_called_once_with(test_payload['instance'],
                                    test_payload['environment'])

    @mock.patch('murano.common.server.get_last_deployment')
    @mock.patch('murano.common.server.session')
    @mock.patch('murano.common.server.models')
    def test_report_notification(self, mock_models, mock_db_session,
                                 mock_last_deployment):
        mock_last_deployment.return_value = mock.MagicMock(
            id='test_deployment_id')

        test_report = {
            'id': 'test_report_id',
            'timestamp': datetime.now().isoformat(),
            'created': None
        }

        server.report_notification(test_report)
        self.assertIsNotNone(test_report['created'])
        mock_models.Status().update.assert_called_once_with(test_report)
        self.assertEqual('test_deployment_id', mock_models.Status().task_id)
        mock_db_session.get_session().add.assert_called_once_with(
            mock_models.Status())

    def test_get_last_deployment(self):
        mock_unit = mock.MagicMock()
        result = server.get_last_deployment(mock_unit, 'test_env_id')
        self.assertEqual(mock_unit.query().filter_by().order_by().first(),
                         result)
        mock_unit.query().filter_by.assert_any_call(
            environment_id='test_env_id')

    def test_service_class(self):
        service = server.Service()
        self.assertIsNone(service.server)

        # Test stop server.
        service.server = mock.MagicMock()
        service.stop(graceful=True)
        service.server.stop.assert_called_once_with()
        service.server.wait.assert_called_once_with()

        # Test reset server.
        service.reset()
        service.server.reset.assert_called_once_with()

    @mock.patch('murano.common.rpc.messaging')
    def test_notification_service_class(self, mock_messaging):
        mock_server = mock.MagicMock()
        mock_messaging.get_notification_listener.return_value = mock_server
        notification_service = server.NotificationService()
        self.assertIsNone(notification_service.server)

        notification_service.start()
        self.assertEqual(1,
                         mock_messaging.get_notification_listener.call_count)
        mock_server.start.assert_called_once_with()

    @mock.patch('murano.common.rpc.messaging')
    def test_api_service_class(self, mock_messaging):
        mock_server = mock.MagicMock()
        mock_messaging.get_rpc_server.return_value = mock_server
        api_service = server.ApiService()

        api_service.start()
        self.assertEqual(1,
                         mock_messaging.get_rpc_server.call_count)
        mock_server.start.assert_called_once_with()



#    Copyright (c) 2014 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import weakref

from murano.dsl.contracts import contracts
from murano.dsl import dsl_types
from murano.dsl import helpers


class Spec(object):
    def __init__(self, declaration, container_type):
        self._container_type = weakref.ref(container_type)
        self._contract = contracts.Contract(
            declaration['Contract'], container_type)
        self._has_default = 'Default' in declaration
        self._default = declaration.get('Default')

    def _get_this_context(self, this):
        executor = helpers.get_executor()
        if isinstance(this, dsl_types.MuranoType):
            return executor.create_object_context(this)
        return executor.create_object_context(
            this.cast(self._container_type()))

    def transform(self, value, this, owner, context, default=None,
                  finalize=True):
        if default is None:
            default = self.default
        if isinstance(this, dsl_types.MuranoTypeReference):
            this = this.type
        if isinstance(this, dsl_types.MuranoType):
            return self._contract.transform(
                value, self._get_this_context(this),
                None, None, default, helpers.get_type(context),
                finalize=finalize)
        else:
            return self._contract.transform(
                value, self._get_this_context(this),
                this, owner, default, helpers.get_type(context),
                finalize=finalize)

    def validate(self, value, this, context, default=None):
        if default is None:
            default = self.default
        return self._contract.validate(
            value, self._get_this_context(this), default,
            helpers.get_type(context))

    def check_type(self, value, context, default=None):
        if default is None:
            default = self.default
        return self._contract.check_type(
            value, context, default, helpers.get_type(context))

    def finalize(self, value, this, context):
        return self._contract.finalize(
            value, self._get_this_context(this), helpers.get_type(context))

    @property
    def default(self):
        return self._default

    @property
    def contract(self):
        return self._contract

    @property
    def has_default(self):
        return self._has_default

    @property
    def declaring_type(self):
        return self._container_type()



#    Copyright (c) 2015 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import base64

from keystoneauth1 import exceptions
from keystoneauth1.identity import v3
from keystoneauth1 import session as ks_session
from oslo_config import cfg
from oslo_log import log
from webob import exc

from murano.common.i18n import _
from murano.common import wsgi

CONF = cfg.CONF
LOG = log.getLogger(__name__)


class ExternalContextMiddleware(wsgi.Middleware):
    def get_keystone_token(self, user, password):
        # TODO(starodubcevna): picking up project_name and auth_url from
        # section related to Cloud Foundry service broker is probably a duct
        # tape and should be rewritten as soon as we get more non-OpenStack
        # services as murano recipients. The same is right for project and user
        # domain names.

        auth_url = CONF.cfapi.auth_url
        if not (auth_url.endswith('v2.0') or auth_url.endswith('v3')):
            auth_url += '/v3'
        elif auth_url.endswith('v2.0'):
            auth_url = auth_url.replace('v2.0', 'v3')
        elif auth_url.endswith('v3'):
            pass
        else:
            LOG.warning('Wrong format for service broker auth url')

        kwargs = {'auth_url': auth_url,
                  'username': user,
                  'password': password,
                  'project_name': CONF.cfapi.tenant,
                  'user_domain_name': CONF.cfapi.user_domain_name,
                  'project_domain_name': CONF.cfapi.project_domain_name}
        password_auth = v3.Password(**kwargs)
        session = ks_session.Session(auth=password_auth)

        self._query_endpoints(password_auth, session)

        return session.get_token()

    def _query_endpoints(self, auth, session):
        if not hasattr(self, '_murano_endpoint'):
            try:
                self._murano_endpoint = auth.get_endpoint(
                    session, 'application-catalog')
            except exceptions.EndpointNotFound:
                pass
        if not hasattr(self, '_glare_endpoint'):
            try:
                self._glare_endpoint = auth.get_endpoint(
                    session, 'artifact')
            except exceptions.EndpointNotFound:
                pass

    def get_endpoints(self):
        return {
            'murano': getattr(self, '_murano_endpoint', None),
            'glare': getattr(self, '_glare_endpoint', None),
        }

    def process_request(self, req):

        """Get keystone token for external request

        This middleware is used for external requests. It get credentials from
        request and try to get keystone token for further authorization.

        :param req: wsgi request object that will be given the context object
        """
        try:
            credentials = base64.b64decode(
                req.headers['Authorization'].split(' ')[1])
            user, password = credentials.decode('utf-8').split(':', 2)
            req.headers['X-Auth-Token'] = self.get_keystone_token(user,
                                                                  password)
            req.endpoints = self.get_endpoints()
        except KeyError:
            msg = _("Authorization required")
            LOG.warning(msg)
            raise exc.HTTPUnauthorized(explanation=msg)
        except exceptions.Unauthorized:
            msg = _("Your credentials are wrong. Please try again")
            LOG.warning(msg)
            raise exc.HTTPUnauthorized(explanation=msg)

    @classmethod
    def factory(cls, global_conf, **local_conf):
        def filter(app):
            return cls(app)
        return filter



#    Copyright (c) 2014 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
import gc
import weakref

from oslo_log import log as logging

from murano.dsl import dsl_types
from murano.dsl import helpers
from murano.dsl import murano_object


LOG = logging.getLogger(__name__)


class ObjectStore(object):
    def __init__(self, executor, parent_store=None, weak_store=True):
        self._parent_store = parent_store
        self._store = weakref.WeakValueDictionary() if weak_store else {}
        self._designer_attributes_store = {}
        self._executor = weakref.ref(executor)
        self._pending_destruction = set()

    @property
    def executor(self):
        return self._executor()

    def get(self, object_id, check_parent_store=True):
        result = self._store.get(object_id)
        if not result and self._parent_store and check_parent_store:
            return self._parent_store.get(object_id, check_parent_store)
        return result

    def has(self, object_id, check_parent_store=True):
        if object_id in self._store:
            return True
        if check_parent_store and self._parent_store:
            return self._parent_store.has(object_id, check_parent_store)
        return False

    def put(self, murano_object, object_id=None):
        self._store[object_id or murano_object.object_id] = murano_object

    def schedule_object_destruction(self, murano_object):
        self._pending_destruction.add(murano_object)
        self._store[murano_object.object_id] = murano_object

    def iterate(self):
        return self._store.keys()

    def remove(self, object_id):
        self._store.pop(object_id)

    def load(self, value, owner, default_type=None,
             scope_type=None, context=None, keep_ids=False,
             bypass_store=False):
        # do the object model load in a temporary object store and copy
        # loaded objects here after that
        model_store = InitializationObjectStore(
            owner, self, keep_ids)
        with helpers.with_object_store(model_store):
            result = model_store.load(
                value, owner, scope_type=scope_type,
                default_type=default_type, context=context)
            for obj_id in model_store.iterate():
                obj = model_store.get(obj_id)
                if obj.initialized:
                    if not bypass_store:
                        self.put(obj)
            return result

    @staticmethod
    def _get_designer_attributes(header):
        return dict((k, v) for k, v in header.items()
                    if str(k).startswith('_'))

    def designer_attributes(self, object_id):
        return self._designer_attributes_store.get(object_id, {})

    @property
    def initializing(self):
        return False

    @property
    def parent_store(self):
        return self._parent_store

    def cleanup(self):
        LOG.debug('Cleaning up orphan objects')
        with helpers.with_object_store(self):
            n = self._collect_garbage()
            LOG.debug('{} orphan objects were destroyed'.format(n))
            return n

    def prepare_finalize(self, used_objects):
        used_objects = set(used_objects) if used_objects else []
        sentenced_objects = [
            obj for obj in self._store.values()
            if obj not in used_objects
        ]
        with helpers.with_object_store(self):
            if sentenced_objects:
                self._pending_destruction.update(sentenced_objects)
                for __ in self._destroy_garbage(sentenced_objects):
                    pass

    def finalize(self):
        with helpers.with_object_store(self):
            for t in self._store.values():
                t.mark_destroyed(True)
        self._pending_destruction.clear()
        self._store.clear()

    def _collect_garbage(self):
        repeat = True
        count = 0
        while repeat:
            repeat = False
            gc.collect()
            for obj in gc.garbage:
                if (isinstance(obj, murano_object.RecyclableMuranoObject) and
                        obj.executor is self._executor()):
                    repeat = True
                    if obj.initialized and not obj.destroyed:
                        self.schedule_object_destruction(obj)
                    else:
                        obj.mark_destroyed(True)
            obj = None
            del gc.garbage[:]
            if self._pending_destruction:
                for obj in self._destroy_garbage(self._pending_destruction):
                    if obj in self._pending_destruction:
                        repeat = True
                        obj.mark_destroyed()
                        self._pending_destruction.remove(obj)
                        count += 1
        return count

    def is_doomed(self, obj):
        return obj.destroyed or obj in self._pending_destruction

    def _destroy_garbage(self, sentenced_objects):
        dd_graph = {}

        # NOTE(starodubcevna): construct a graph which looks like:
        # {
        #   obj1: [subscriber1, subscriber2],
        #   obj2: [subscriber2, subscriber3]
        # }
        for obj in sentenced_objects:
            obj_subscribers = [obj.owner]

            for dd in obj.destruction_dependencies:
                subscriber = dd['subscriber']
                if subscriber:
                    subscriber = subscriber()
                if subscriber and subscriber not in obj_subscribers:
                    obj_subscribers.append(subscriber)

            dd_graph[obj] = obj_subscribers

        def topological(graph):
            """Topological sort implementation

            This implementation will work even if we have cycle dependencies,
            e.g. [a->b, b->c, c->a]. In this case the order of deletion will be
            undefined and it's okay.
            """

            visited = collections.defaultdict(int)
            indexes = collections.defaultdict(int)

            def dfs(obj):
                visited[obj] += 1
                subscribers = graph.get(obj)
                if subscribers is not None:
                    m = 0
                    for i, subscriber in enumerate(subscribers):
                        if i == 0 or not visited[subscriber]:
                            for t in dfs(subscriber):
                                yield t
                        m = max(m, indexes[subscriber])

                    if visited.get(obj, 0) <= 2:
                        visited[obj] += 1
                        indexes[obj] = m + 1
                        yield obj, m + 1

            for i, obj in enumerate(graph.keys()):
                if not visited[obj]:
                    for t in dfs(obj):
                        yield t

            visited.clear()
            indexes.clear()

        order = collections.defaultdict(list)
        for obj, index in topological(dd_graph):
            order[index].append(obj)

        for key in sorted(order):
            group = order[key]
            self.executor.signal_destruction_dependencies(*group)

        for key in sorted(order, reverse=True):
            group = order[key]
            self.executor.destroy_objects(*group)
            for t in group:
                yield t


# Temporary ObjectStore to load object graphs. Does 2-phase load
# and maintains internal state on what phase is currently running
# as well as objects that are in the middle of initialization.
# Required in order to isolate semi-initialized objects from regular
# objects in main ObjectStore and internal state between graph loads
# in different threads. Once the load is done all objects are copied
# to the parent ObjectStore
class InitializationObjectStore(ObjectStore):
    def __init__(self, root_owner, parent_store, keep_ids):
        super(InitializationObjectStore, self).__init__(
            parent_store.executor, parent_store, weak_store=False)
        self._initializing = False
        self._root_owner = root_owner
        self._keep_ids = keep_ids
        self._initializers = []

    @property
    def initializing(self):
        return self._initializing

    def load(self, value, owner, default_type=None,
             scope_type=None, context=None, **kwargs):
        parsed = helpers.parse_object_definition(value, scope_type, context)
        if not parsed:
            raise ValueError('Invalid object representation format')

        if owner is self._root_owner:
            self._initializing = True

        class_obj = parsed['type'] or default_type
        if not class_obj:
            raise ValueError(
                'Invalid object representation: '
                'no type information was provided')
        if isinstance(class_obj, dsl_types.MuranoTypeReference):
            class_obj = class_obj.type
        object_id = parsed['id']
        is_tmp_object = (object_id is None and
                         owner is not self._root_owner and
                         self._initializing)
        obj = None if object_id is None else self.get(
            object_id, self._keep_ids)
        if not obj:
            if is_tmp_object or helpers.is_objects_dry_run_mode():
                mo_type = murano_object.MuranoObject
            else:
                mo_type = murano_object.RecyclableMuranoObject
            obj = mo_type(
                class_obj, owner,
                name=parsed['name'],
                metadata=parsed['metadata'],
                object_id=object_id if self._keep_ids else None)
            obj.load_dependencies(parsed['dependencies'])
            if parsed['destroyed']:
                obj.mark_destroyed()
            self.put(obj, object_id or obj.object_id)

            system_value = ObjectStore._get_designer_attributes(
                parsed['extra'])
            self._designer_attributes_store[object_id] = system_value

        if context is None:
            context = self.executor.create_object_context(obj)

        def run_initialize():
            self._initializers.extend(
                obj.initialize(context, parsed['properties']))

        run_initialize()
        if owner is self._root_owner:
            self._initializing = False
            run_initialize()

        if owner is self._root_owner:
            with helpers.with_object_store(self.parent_store):
                for fn in self._initializers:
                    fn()

        return obj



# Copyright (c) 2016 AT&T Corp
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os
import random
import semantic_version
import shutil
import string
import tempfile
from unittest import mock

from murano.packages import exceptions
from murano.packages import package_base
import murano.tests.unit.base as test_base


class TestPackageBase(test_base.MuranoTestCase):

    @classmethod
    def setUpClass(cls):
        super(TestPackageBase, cls).setUpClass()
        package_base.PackageBase.__abstractmethods__ = set()
        cls.source_directory = tempfile.mkdtemp(dir=tempfile.tempdir)
        cls.version = semantic_version.Version.coerce('1.2.3')
        cls.mock_manifest = {
            'Name': 'mock_display_name',
            'FullName': 'mock_full_name',
            'Type': 'Application',
            'Version': '1.2.3',
            'Description': 'test_description',
            'Author': 'test_author',
            'Supplier': 'test_supplier',
            'Tags': ['tag1', 'tag2', 'tag3'],
            'Logo': None
        }
        cls.package_base = package_base.PackageBase('test_format',
                                                    'test_runtime_version',
                                                    cls.source_directory,
                                                    cls.mock_manifest)

    @classmethod
    def tearDownClass(cls):
        if os.path.isdir(cls.source_directory):
            shutil.rmtree(cls.source_directory)

    def test_create_package_base_without_full_name(self):
        with self.assertRaisesRegex(exceptions.PackageFormatError,
                                    'FullName is not specified'):
            package_base.PackageBase('test_format',
                                     'test_runtime_version',
                                     'test_source_directory',
                                     manifest={'FullName': None})

    def test_create_package_base_with_invalid_full_name(self):
        full_names = ['.invalid_name_1', 'invalid..name..2', 'invalid name 3']
        for full_name in full_names:
            expected_error_message = 'Invalid FullName {0}'.format(full_name)
            with self.assertRaisesRegex(exceptions.PackageFormatError,
                                        expected_error_message):
                package_base.PackageBase('test_format',
                                         'test_runtime_version',
                                         'test_source_directory',
                                         manifest={'FullName': full_name})

    def test_create_package_base_with_invalid_type(self):
        package_type = 'Invalid'
        with self.assertRaisesRegex(exceptions.PackageFormatError,
                                    'Invalid package Type {0}'
                                    .format(package_type)):
            package_base.PackageBase('test_format',
                                     'test_runtime_version',
                                     'test_source_directory',
                                     manifest={'FullName': 'mock_full_name',
                                               'Type': package_type})

    def test_requirements_negative(self):
        with self.assertRaisesRegex(NotImplementedError, None):
            self.package_base.requirements

    def test_classes_negative(self):
        with self.assertRaisesRegex(NotImplementedError, None):
            self.package_base.classes

    def test_get_class_negative(self):
        with self.assertRaisesRegex(NotImplementedError, None):
            self.package_base.get_class(None)

    def test_ui_negative(self):
        with self.assertRaisesRegex(NotImplementedError, None):
            self.package_base.ui

    def test_full_name(self):
        self.assertEqual(self.mock_manifest['FullName'],
                         self.package_base.full_name)

    def test_source_directory(self):
        self.assertEqual(self.source_directory,
                         self.package_base.source_directory)

    def test_version(self):
        self.assertEqual(self.version,
                         self.package_base.version)

    def test_package_type(self):
        self.assertEqual(self.mock_manifest['Type'],
                         self.package_base.package_type)

    def test_display_name(self):
        self.assertEqual(self.mock_manifest['Name'],
                         self.package_base.display_name)

    def test_description(self):
        self.assertEqual(self.mock_manifest['Description'],
                         self.package_base.description)

    def test_author(self):
        self.assertEqual(self.mock_manifest['Author'],
                         self.package_base.author)

    def test_supplier(self):
        self.assertEqual(self.mock_manifest['Supplier'],
                         self.package_base.supplier)

    def test_tags(self):
        self.assertEqual(self.mock_manifest['Tags'],
                         self.package_base.tags)

    def test_logo_without_file_name(self):
        self.assertIsNone(self.package_base.logo)

    def test_logo_with_invalid_logo_path(self):
        expected_error_message = 'Unable to load logo'
        self.package_base._logo = ''.join(random.choice(string.ascii_letters)
                                          for _ in range(10))
        with self.assertRaisesRegex(exceptions.PackageLoadError,
                                    expected_error_message):
            self.package_base.logo
        self.package_base._logo = self.mock_manifest['Logo']

    @mock.patch('murano.packages.package_base.imghdr',
                what=mock.MagicMock(return_value='xyz'))
    def test_load_image_with_invalid_extension(self, mock_imghdr):
        expected_error_message = 'Unsupported Format.'
        with self.assertRaisesRegex(exceptions.PackageLoadError,
                                    expected_error_message):
            self.package_base._load_image('logo.xyz', 'logo.xyz', 'logo')
            full_path = os.path.join(self.package_base._source_directory,
                                     'logo.xyz')
            mock_imghdr.what.assert_called_once_with(full_path)

    @mock.patch('murano.packages.package_base.imghdr',
                what=mock.MagicMock(return_value='png'))
    @mock.patch('murano.packages.package_base.os')
    def test_load_image_with_oversized_image(self, mock_os, mock_imghdr):
        mock_os.stat.return_value = mock.MagicMock(st_size=5000 * 1024)
        mock_os.isfile = mock.MagicMock(return_value=True)
        expected_error_message = 'Max allowed size is {0}'.format(500 * 1024)
        with self.assertRaisesRegex(exceptions.PackageLoadError,
                                    expected_error_message):
            self.package_base._load_image('logo.xyz', 'logo.xyz', 'logo')

    def test_meta(self):
        self.assertIsNone(self.package_base.meta)

    def test_get_resource(self):
        test_name = 'test_resource_name'
        expected_dir = os.path.join(self.source_directory, 'Resources',
                                    test_name)
        self.assertEqual(expected_dir, self.package_base.get_resource(
            test_name))



# Copyright (c) 2014 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from murano.tests.unit.dsl.foundation import object_model as om
from murano.tests.unit.dsl.foundation import test_case


class TestMultipleInheritance(test_case.DslTestCase):
    def setUp(self):
        super(TestMultipleInheritance, self).setUp()
        self._multi_derived = om.Object(
            'DerivedFrom2Classes',
            rootProperty='ROOT')
        model = om.Object(
            'SampleClass3',
            multiClassProperty=self._multi_derived
        )
        self._runner = self.new_runner(model)

    def test_multi_class_contract(self):
        self._runner.testMultiContract()
        self.assertEqual(
            ['ParentClass1::method1', 'ParentClass2::method2'],
            self.traces)

    def test_root_method_resolution(self):
        self._runner.on(self._multi_derived).testRootMethod()
        self.assertEqual(
            ['CommonParent::testRootMethod', 'ROOT'],
            self.traces)

    def test_property_accessible_on_several_paths(self):
        self.assertEqual(
            'ROOT',
            self._runner.testPropertyAccessibleOnSeveralPaths())

    def test_specialized_mixin_override(self):
        self._runner.on(self._multi_derived).testMixinOverride()
        self.assertEqual(
            ['ParentClass2::virtualMethod', '-',
             'CommonParent::virtualMethod', '-',
             'CommonParent::virtualMethod', '-',
             'ParentClass2::virtualMethod'],
            self.traces)

    def test_super(self):
        self._runner.on(self._multi_derived).testSuper()
        self.assertCountEqual(
            ['CommonParent::virtualMethod', 'ParentClass2::virtualMethod',
             'CommonParent::virtualMethod', 'ParentClass2::virtualMethod'],
            self.traces)

    def test_psuper(self):
        self._runner.on(self._multi_derived).testPsuper()
        self.assertCountEqual(
            ['CommonParent::virtualMethod', 'ParentClass2::virtualMethod',
             'CommonParent::virtualMethod', 'ParentClass2::virtualMethod'],
            self.traces)



#    Copyright (c) 2014 Hewlett-Packard Development Company, L.P.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

# Based on designate/policy.py

from oslo_config import cfg
from oslo_log import log as logging
from oslo_policy import opts
from oslo_policy import policy
from webob import exc as exceptions

from murano.common import policies

LOG = logging.getLogger(__name__)

CONF = cfg.CONF

_ENFORCER = None

# TODO(gmann): Remove setting the default value of config policy_file
# once oslo_policy change the default value to 'policy.yaml'.
# https://github.com/openstack/oslo.policy/blob/a626ad12fe5a3abd49d70e3e5b95589d279ab578/oslo_policy/opts.py#L49
DEFAULT_POLICY_FILE = 'policy.yaml'
opts.set_defaults(CONF, DEFAULT_POLICY_FILE)


def reset():
    global _ENFORCER
    if _ENFORCER:
        _ENFORCER.clear()
    _ENFORCER = None


def init(use_conf=True):
    global _ENFORCER
    if not _ENFORCER:
        LOG.debug("Enforcer is not present, recreating.")
        _ENFORCER = policy.Enforcer(CONF, use_conf=use_conf)
        register_rules(_ENFORCER)


def set_rules(data, default_rule=None, overwrite=True):
    default_rule = default_rule or cfg.CONF.policy_default_rule
    if not _ENFORCER:
        LOG.debug("Enforcer not present, recreating at rules stage.")
        init()

    if default_rule:
        _ENFORCER.default_rule = default_rule

    LOG.debug("Loading rules {rules}, default: {def_rule}, overwrite: "
              "{overwrite}".format(rules=data, def_rule=default_rule,
                                   overwrite=overwrite))

    if isinstance(data, dict):
        rules = policy.Rules.from_dict(data, default_rule)
    else:
        rules = policy.Rules.load_json(data, default_rule)

    _ENFORCER.set_rules(rules, overwrite=overwrite)


def check(rule, ctxt, target=None, do_raise=True, exc=None):
    """Verify that the rule is valid on the target in this context.

    :param rule: String representing the action to be checked, which should
                 be colon-separated for clarity.
    :param ctxt: Request context from which user credentials are extracted.
    :param target: Dictionary representing the object of the action for object
                   creation; this should be a dictionary representing the
                   location of the object, e.g. {'environment_id':
                   object.environment_id}
    :param do_raise: Whether to raise an exception or not if the check fails.
    :param exc: Class of the exception to raise if the check fails.
    :raises exceptions.HTTPForbidden: If verification fails. Or if 'exc' is
                                      specified it will raise an exception of
                                      that type.
    """
    init()

    if target is None:
        target = {}
    creds = ctxt.to_dict()
    if not exc:
        exc = exceptions.HTTPForbidden

    try:
        result = _ENFORCER.enforce(rule, target, creds, do_raise, exc)
    except Exception:
        result = False
        raise
    else:
        return result
    finally:
        if result:
            LOG.debug("Policy check succeeded for rule {rule} on target "
                      "{target}".format(rule=rule, target=repr(target)))
        else:
            LOG.debug("Policy check failed for rule {rule} on target: "
                      "{target}".format(rule=rule, target=repr(target)))


def check_is_admin(context):
    """Check if the given context is associated with an admin role.

       :param context: Murano request context
       :returns: A non-False value if context role is admin.
    """
    return check('context_is_admin', context, context.to_dict(),
                 do_raise=False)


def register_rules(enforcer):
    enforcer.register_defaults(policies.list_rules())



# Copyright (c) 2015 OpenStack Foundation.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
# http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.


import json

import eventlet
try:
    from mistralclient.api import client as mistralcli
except ImportError:
    mistralcli = None
from oslo_config import cfg
from oslo_log import log as logging

from murano.common import auth_utils
from murano.dsl import dsl
from murano.dsl import session_local_storage

CONF = cfg.CONF
LOG = logging.getLogger(__name__)


class MistralError(Exception):
    pass


@dsl.name('io.murano.system.MistralClient')
class MistralClient(object):
    def __init__(self, this, region_name=None):
        self._owner = this.find_owner('io.murano.Environment')
        self._region_name = region_name

    @property
    def _client(self):
        region = self._region_name or (
            None if self._owner is None else self._owner['region'])
        return self._create_client(region)

    @staticmethod
    @session_local_storage.execution_session_memoize
    def _create_client(region):
        if not mistralcli:
            LOG.warning("Mistral client is not available")
            raise ImportError("Import mistralcliet error")

        mistral_settings = CONF.mistral

        endpoint_type = mistral_settings.endpoint_type
        service_type = mistral_settings.service_type
        session = auth_utils.get_client_session()

        mistral_url = mistral_settings.url or session.get_endpoint(
            service_type=service_type,
            endpoint_type=endpoint_type,
            region_name=region)
        auth_ref = session.auth.get_access(session)

        # TODO(gyurco): use auth_utils.get_session_client_parameters
        return mistralcli.client(
            mistral_url=mistral_url,
            project_id=auth_ref.project_id,
            endpoint_type=endpoint_type,
            service_type=service_type,
            auth_token=auth_ref.auth_token,
            user_id=auth_ref.user_id,
            insecure=mistral_settings.insecure,
            cacert=mistral_settings.cafile
        )

    def upload(self, definition):
        self._client.workflows.create(definition)

    def run(self, name, timeout=600, inputs=None, params=None):
        execution = self._client.executions.create(
            workflow_identifier=name, workflow_input=inputs, params=params)
        # For the fire and forget functionality - when we do not want to wait
        # for the result of the run.
        if timeout == 0:
            return execution.id

        state = execution.state
        try:
            # While the workflow is running we continue to wait until timeout.
            with eventlet.timeout.Timeout(timeout):
                while state not in ('ERROR', 'SUCCESS'):
                    eventlet.sleep(2)
                    execution = self._client.executions.get(execution.id)
                    state = execution.state
        except eventlet.timeout.Timeout:
            error_message = (
                'Mistral run timed out. Execution id: {0}.').format(
                execution.id)
            raise MistralError(error_message)

        if state == 'ERROR':
            error_message = ('Mistral execution completed with ERROR.'
                             ' Execution id: {0}. Output: {1}').format(
                execution.id, execution.output)
            raise MistralError(error_message)

        # Load the JSON we got from Mistral client to dictionary.
        output = json.loads(execution.output)

        # Clean the returned dictionary from unnecessary data.
        # We want to keep only flow level outputs.
        output.pop('openstack', None)
        output.pop('__execution', None)
        output.pop('task', None)

        return output



#    Copyright (c) 2014 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

from murano.common.helpers import path
from murano.packages import exceptions
from murano.packages import package_base


class MuranoPlPackage(package_base.PackageBase):
    def __init__(self, format_name, runtime_version, source_directory,
                 manifest):
        super(MuranoPlPackage, self).__init__(
            format_name, runtime_version, source_directory, manifest)
        self._classes = manifest.get('Classes')
        self._ui_file = manifest.get('UI', 'ui.yaml')
        self._requirements = manifest.get('Require') or {}
        self._meta = manifest.get('Meta')

    @property
    def classes(self):
        return self._classes.keys()

    @property
    def ui(self):
        full_path = path.secure_join(
            self._source_directory, 'UI', self._ui_file)
        if not os.path.isfile(full_path):
            return None
        with open(full_path, 'rb') as stream:
            return stream.read()

    @property
    def requirements(self):
        return self._requirements

    def get_class(self, name):
        if name not in self._classes:
            raise exceptions.PackageClassLoadError(
                name, 'Class not defined in package ' + self.full_name)
        def_file = self._classes[name]
        full_path = path.secure_join(
            self._source_directory, 'Classes', def_file)
        if not os.path.isfile(full_path):
            raise exceptions.PackageClassLoadError(
                name, 'File with class definition not found')
        with open(full_path, 'rb') as stream:
            return stream.read(), full_path

    @property
    def meta(self):
        return self._meta



# Copyright (c) 2014 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from testtools import matchers

from murano.dsl import dsl
from murano.dsl import serializer
from murano.tests.unit.dsl.foundation import object_model as om
from murano.tests.unit.dsl.foundation import test_case


class TestConstruction(test_case.DslTestCase):
    def setUp(self):
        super(TestConstruction, self).setUp()
        self._runner = self.new_runner(om.Object('CreatingClass'))

    def test_new(self):
        self._runner.testNew()
        self.assertEqual(
            ['CreatingClass::.init', 'CreatedClass1::.init',
             'string', 'STRING', 123],
            self.traces)

    def test_new_with_ownership(self):
        obj = serializer.serialize(self._runner.testNewWithOwnership(),
                                   self._runner.executor, allow_refs=False)
        self.assertEqual('STRING', obj.get('property1'))
        self.assertIsNotNone('string', obj.get('xxx'))
        self.assertEqual('STR', obj['xxx'].get('property1'))
        self.assertEqual('QQQ', obj['xxx']['?'].get('name'))

    def test_new_with_dict(self):
        self._runner.testNewWithDict()
        self.assertEqual(
            ['CreatingClass::.init', 'CreatedClass1::.init',
             'string', 'STRING', 123],
            self.traces)

    def test_model_load(self):
        res = self._runner.testLoadCompexModel()
        for i in range(3):
            self.assertThat(res[i], matchers.Not(matchers.Contains('node')))
        self.assertEqual(self._runner.root.object_id, res[3])
        self.assertEqual(
            [
                'rootNode',
                ['childNode1', 'childNode2', 'childNode2'],
                True, True, True, True, True,
                'rootNode', 'childNode2', 'childNode1'
            ], res[4:])

    def test_single_contract_instantiation(self):
        self._runner.testSingleContractInstantiation()
        self.assertEqual(1, self.traces.count('ConstructionSample::init'))

    def test_nested_new_loads_in_separate_store(self):
        res = self._runner.testNestedNewLoadsInSeparateStore()
        self.assertIsInstance(res, dsl.MuranoObjectInterface)

    def test_reference_access_from_init(self):
        self._runner.testReferenceAccessFromInit()
        self.assertEqual(2, self.traces.count('childNode'))



# Copyright 2017 AT&T Corporation.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import itertools

from murano.common.policies import action
from murano.common.policies import base
from murano.common.policies import category
from murano.common.policies import deployment
from murano.common.policies import env_template
from murano.common.policies import environment
from murano.common.policies import package


def list_rules():
    return itertools.chain(
        base.list_rules(),
        action.list_rules(),
        category.list_rules(),
        deployment.list_rules(),
        environment.list_rules(),
        env_template.list_rules(),
        package.list_rules()
    )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from io import StringIO
import os
import sys
from unittest import mock

import fixtures
from oslo_config import cfg
from oslo_log import log as logging
from oslo_utils import importutils
import testtools

from murano.cmd import test_runner
from murano import version

CONF = cfg.CONF
logging.register_options(CONF)
logging.setup(CONF, 'murano')


class TestCaseShell(testtools.TestCase):
    def setUp(self):
        super(TestCaseShell, self).setUp()
        self.auth_params = {'username': 'test',
                            'password': 'test',
                            'project_name': 'test',
                            'auth_url': 'http://localhost:5000'}
        self.args = ['test-runner.py']
        for k, v in self.auth_params.items():
            k = '--os-' + k.replace('_', '-')
            self.args.extend([k, v])

        self.useFixture(fixtures.MonkeyPatch('keystoneclient.v3.client.Client',
                                             mock.MagicMock))
        dirs = [os.path.dirname(__file__),
                os.path.join(os.path.dirname(__file__), os.pardir, os.pardir,
                             os.pardir, os.pardir, os.pardir, 'meta')]
        self.override_config('load_packages_from', dirs, 'engine')

    def tearDown(self):
        super(TestCaseShell, self).tearDown()
        CONF.clear()

    def override_config(self, name, override, group=None):
        CONF.set_override(name, override, group)
        CONF.set_override('use_stderr', True)
        self.addCleanup(CONF.clear_override, name, group)

    def shell(self, cmd_args=None, exitcode=0):
        stdout = StringIO()
        stderr = StringIO()
        args = self.args
        if cmd_args:
            cmd_args = cmd_args.split()
            args.extend(cmd_args)
        with mock.patch.object(sys, 'stdout', stdout):
            with mock.patch.object(sys, 'stderr', stderr):
                with mock.patch.object(sys, 'argv', args):
                    result = self.assertRaises(SystemExit, test_runner.main)
                    self.assertEqual(result.code, exitcode,
                                     'Command finished with error.')
        stdout = stdout.getvalue()
        stderr = stderr.getvalue()
        return (stdout, stderr)

    def test_help(self):
        stdout, _ = self.shell('--help')
        usage = """usage: murano-test-runner [-h] [--config-file CONFIG_FILE]
                          [--os-auth-url OS_AUTH_URL]
                          [--os-username OS_USERNAME]
                          [--os-password OS_PASSWORD]
                          [--os-project-name OS_PROJECT_NAME]
                          [-l [</path1, /path2> [</path1, /path2> ...]]] [-v]
                          [--version]
                          <PACKAGE_FQN>
                          [<testMethod1, className.testMethod2> [<testMethod1, className.testMethod2"""  # noqa
        self.assertIn(usage, stdout)

    def test_version(self):
        stdout, stderr = self.shell('--version')
        output = stdout
        self.assertIn(version.version_string, output)

    @mock.patch.object(test_runner, 'LOG')
    def test_increase_verbosity(self, mock_log):
        self.shell('io.murano.test.MyTest1 -v')
        mock_log.logger.setLevel.assert_called_with(logging.DEBUG)

    @mock.patch('keystoneclient.v3.client.Client')
    def test_os_params_replaces_config(self, mock_client):
        # Load keystone configuration parameters from config
        importutils.import_module('keystonemiddleware.auth_token')
        self.override_config('admin_user', 'new_value', 'keystone_authtoken')

        self.shell('io.murano.test.MyTest1 io.murano.test.MyTest2')

        mock_client.assert_has_calls([mock.call(**self.auth_params)])

    def test_package_all_tests(self):
        _, stderr = self.shell('io.murano.test.MyTest1 -v')
        # NOTE(efedorova): May be, there is a problem with test-runner, since
        # all logs are passed to stderr
        self.assertIn('Test io.murano.test.MyTest1.testSimple1 successful',
                      stderr)
        self.assertIn('Test io.murano.test.MyTest1.testSimple2 successful',
                      stderr)
        self.assertIn('Test io.murano.test.MyTest2.testSimple1 successful',
                      stderr)
        self.assertIn('Test io.murano.test.MyTest2.testSimple2 successful',
                      stderr)
        self.assertNotIn('thisIsNotAtestMethod', stderr)

    def test_package_by_class(self):
        _, stderr = self.shell(
            'io.murano.test.MyTest1 io.murano.test.MyTest2 -v')

        self.assertNotIn('Test io.murano.test.MyTest1.testSimple1 successful',
                         stderr)
        self.assertNotIn('Test io.murano.test.MyTest1.testSimple2 successful',
                         stderr)
        self.assertIn('Test io.murano.test.MyTest2.testSimple1 successful',
                      stderr)
        self.assertIn('Test io.murano.test.MyTest2.testSimple2 successful',
                      stderr)

    def test_package_by_test_name(self):
        _, stderr = self.shell(
            'io.murano.test.MyTest1 testSimple1 -v')

        self.assertIn('Test io.murano.test.MyTest1.testSimple1 successful',
                      stderr)
        self.assertNotIn('Test io.murano.test.MyTest1.testSimple2 successful',
                         stderr)
        self.assertIn('Test io.murano.test.MyTest2.testSimple1 successful',
                      stderr)
        self.assertNotIn('Test io.murano.test.MyTest2.testSimple2 successful',
                         stderr)

    def test_package_by_test_and_class_name(self):
        _, stderr = self.shell(
            'io.murano.test.MyTest1 io.murano.test.MyTest2.testSimple1 -v')

        self.assertNotIn('Test io.murano.test.MyTest1.testSimple1 successful',
                         stderr)
        self.assertNotIn('Test io.murano.test.MyTest1.testSimple2 successful',
                         stderr)
        self.assertIn('Test io.murano.test.MyTest2.testSimple1 successful',
                      stderr)
        self.assertNotIn('Test io.murano.test.MyTest2.testSimple2 successful',
                         stderr)

    def test_service_methods(self):
        _, stderr = self.shell(
            'io.murano.test.MyTest1 io.murano.test.MyTest1.testSimple1 -v')
        self.assertIn('Executing: io.murano.test.MyTest1.setUp', stderr)
        self.assertIn('Executing: io.murano.test.MyTest1.tearDown', stderr)

    def test_package_is_not_provided(self):
        _, stderr = self.shell(exitcode=2)
        err = 'the following arguments are required: '
        self.assertIn('murano-test-runner: error: %s' % err, stderr)

    def test_wrong_parent(self):
        _, stderr = self.shell(
            'io.murano.test.MyTest1 io.murano.test.MyTest3 -v', exitcode=1)
        self.assertIn('Class io.murano.test.MyTest3 is not inherited from'
                      ' io.murano.test.TestFixture. Skipping it.', stderr)
        self.assertIn('No tests found for execution.', stderr)



# Copyright 2017 AT&T Corporation.
# All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

from oslo_policy import policy

from murano.common.policies import base

environment_policies = [
    policy.DocumentedRuleDefault(
        name='list_environments',
        check_str=base.RULE_DEFAULT,
        description='List environments in a project.',
        operations=[{'path': '/v1/environments',
                     'method': 'GET'}]),
    policy.DocumentedRuleDefault(
        name='list_environments_all_tenants',
        check_str=base.RULE_ADMIN_API,
        description='List environments across all projects.',
        operations=[{'path': '/v1/environments?all_tenants=true',
                     'method': 'GET'}]),
    policy.DocumentedRuleDefault(
        name='show_environment',
        check_str=base.RULE_DEFAULT,
        description='Show details for an environment or shows the environment '
                    'model.',
        operations=[{'path': '/v1/environments/{environment_id}',
                     'method': 'GET'},
                    {'path': '/v1/environments/{environment_id}/model',
                     'method': 'GET'}]),
    policy.DocumentedRuleDefault(
        name='update_environment',
        check_str=base.RULE_DEFAULT,
        description='Update or rename an environment.',
        operations=[{'path': '/v1/environments/{environment_id}',
                     'method': 'PUT'},
                    {'path': '/v1/environments/{environment_id}/model',
                     'method': 'PATCH'}]),
    policy.DocumentedRuleDefault(
        name='create_environment',
        check_str=base.RULE_DEFAULT,
        description='Create an environment or create an environment and '
                    'session from an environment template.',
        operations=[
            {'path': '/v1/environments/{environment_id}',
             'method': 'POST'},
            {'path': '/v1/templates/{env_template_id}/create-environment',
             'method': 'POST'}]),
    policy.DocumentedRuleDefault(
        name='delete_environment',
        check_str=base.RULE_DEFAULT,
        description='Delete an environment.',
        operations=[{'path': '/v1/environments/{environment_id}',
                     'method': 'DELETE'}])
]


def list_rules():
    return environment_policies



# Copyright (c) 2016 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from murano.common import auth_utils
from murano.dsl import dsl
from murano.dsl import helpers


@dsl.name('io.murano.Project')
class Project(object):
    @classmethod
    def get_current(cls):
        fields = auth_utils.get_project(
            helpers.get_execution_session().project_id)
        return cls._to_object(fields)

    @classmethod
    def get_environment_owner(cls):
        fields = auth_utils.get_project(
            helpers.get_execution_session().environment_owner_project_id)
        return cls._to_object(fields)

    @staticmethod
    def _to_object(fields):
        for field in ('links', 'parent_id', 'enabled'):
            fields.pop(field, None)
        obj_def = {
            'id': fields.pop('id'),
            'name': fields.pop('name'),
            'domain': fields.pop('domain_id', 'Default'),
            'description': fields.pop('description', None),
            'extra': fields
        }
        return dsl.new(obj_def)



# Copyright 2016 AT&T Corp
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import webob

from murano.api.middleware import context
from murano.tests.unit import base

from oslo_config import cfg

CONF = cfg.CONF


class MiddlewareContextTest(base.MuranoTestCase):

    def test_middleware_context_default(self):
        middleware = context.ContextMiddleware(None)
        request_headers = {
            'X-Roles': 'admin',
            'X-User-Id': "",
            'X-Tenant-Id': "",
            'X-Configuration-Session': "",
        }
        request = webob.Request.blank('/environments',
                                      headers=request_headers)
        self.assertFalse(hasattr(request, 'context'))
        middleware.process_request(request)
        self.assertTrue(hasattr(request, 'context'))

    def test_factory_returns_filter(self):
        middleware = context.ContextMiddleware(None)
        result = middleware.factory(CONF)
        self.assertIsNotNone(result)



#    Copyright (c) 2015 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

_FATAL_EXCEPTION_FORMAT_ERRORS = False

# Exceptions from openstack-common


class Error(Exception):
    def __init__(self, message=None):
        super(Error, self).__init__(message)


class OpenstackException(Exception):
    """Base Exception class.

    To correctly use this class, inherit from it and define
    a 'msg_fmt' property. That message will get printf'd
    with the keyword arguments provided to the constructor.
    """
    msg_fmt = "An unknown exception occurred"

    def __init__(self, **kwargs):
        try:
            self._error_string = self.msg_fmt % kwargs

        except Exception:
            if _FATAL_EXCEPTION_FORMAT_ERRORS:
                raise
            else:
                # at least get the io.murano message out if something happened
                self._error_string = self.msg_fmt

    def __str__(self):
        return self._error_string


class UnsupportedContentType(OpenstackException):
    msg_fmt = "Unsupported content type %(content_type)s"


class NotAcceptableContentType(OpenstackException):
    msg_fmt = ("Response with content type %(content_type)s "
               "expected but can not be provided")


class MalformedRequestBody(OpenstackException):
    msg_fmt = "Malformed message body: %(reason)s"

# Murano exceptions


class TimeoutException(Exception):
    pass


class PolicyViolationException(Exception):
    pass


class RouterInfoException(Exception):
    pass



# Copyright (c) 2014 Mirantis Inc.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import re
from unittest import mock

import semantic_version
import yaql
from yaql.language import exceptions
from yaql.language import utils

import murano.dsl.helpers as helpers
import murano.dsl.namespace_resolver as ns_resolver
import murano.dsl.yaql_expression as yaql_expression
from murano.tests.unit import base

ROOT_CLASS = 'io.murano.Object'


class TestNamespaceResolving(base.MuranoTestCase):
    def test_fails_w_empty_name(self):
        resolver = ns_resolver.NamespaceResolver({'=': 'com.example.murano'})

        self.assertRaises(ValueError, resolver.resolve_name, None)

    def test_fails_w_unknown_prefix(self):
        resolver = ns_resolver.NamespaceResolver({'=': 'com.example.murano'})
        name = 'unknown_prefix:example.murano'

        self.assertRaises(KeyError, resolver.resolve_name, name)

    def test_fails_w_prefix_wo_name(self):
        resolver = ns_resolver.NamespaceResolver({'=': 'com.example.murano'})
        name = 'sys:'

        self.assertRaises(ValueError, resolver.resolve_name, name)

    def test_fails_w_excessive_prefix(self):
        ns = {'sys': 'com.example.murano.system'}
        resolver = ns_resolver.NamespaceResolver(ns)
        invalid_name = 'sys:excessive_ns:muranoResource'

        self.assertRaises(ValueError, resolver.resolve_name, invalid_name)

    def test_empty_prefix_is_default(self):
        resolver = ns_resolver.NamespaceResolver({'=': 'com.example.murano'})
        # name without prefix delimiter
        name = 'some.arbitrary.name'

        resolved_name = resolver.resolve_name(':' + name)

        self.assertEqual(
            'com.example.murano.some.arbitrary.name', resolved_name)

    def test_resolves_specified_ns_prefix(self):
        ns = {'sys': 'com.example.murano.system'}
        resolver = ns_resolver.NamespaceResolver(ns)
        short_name, full_name = 'sys:File', 'com.example.murano.system.File'

        resolved_name = resolver.resolve_name(short_name)

        self.assertEqual(full_name, resolved_name)

    def test_resolves_current_ns(self):
        resolver = ns_resolver.NamespaceResolver({'=': 'com.example.murano'})
        short_name, full_name = 'Resource', 'com.example.murano.Resource'

        resolved_name = resolver.resolve_name(short_name)

        self.assertEqual(full_name, resolved_name)

    def test_resolves_w_empty_namespaces(self):
        resolver = ns_resolver.NamespaceResolver({})

        resolved_name = resolver.resolve_name('Resource')

        self.assertEqual('Resource', resolved_name)


class TestHelperFunctions(base.MuranoTestCase):
    def test_generate_id(self):
        generated_id = helpers.generate_id()

        self.assertTrue(re.match(r'[a-z0-9]{32}', generated_id))

    def test_evaluate(self):
        yaql_value = mock.Mock(yaql_expression.YaqlExpression,
                               return_value='atom')
        complex_value = {yaql_value: ['some', (1, yaql_value), 'hi!'],
                         'sample': [yaql_value, range(5)]}
        complex_literal = utils.FrozenDict({
            'atom': ('some', (1, 'atom'), 'hi!'),
            'sample': ('atom', (0, 1, 2, 3, 4))
        })
        context = yaql.create_context()
        evaluated_value = helpers.evaluate(yaql_value, context)
        evaluated_complex_value = helpers.evaluate(complex_value, context)

        self.assertEqual('atom', evaluated_value)
        self.assertEqual(complex_literal, evaluated_complex_value)


class TestYaqlExpression(base.MuranoTestCase):
    def setUp(self):
        self._version = semantic_version.Version.coerce('1.0')
        super(TestYaqlExpression, self).setUp()

    def test_expression(self):
        yaql_expr = yaql_expression.YaqlExpression('string', self._version)

        self.assertEqual('string', yaql_expr.expression)

    def test_evaluate_calls(self):
        string = 'string'
        expected_calls = [mock.call(string, self._version),
                          mock.call().evaluate(context=None)]

        with mock.patch('murano.dsl.yaql_integration.parse') as mock_parse:
            yaql_expr = yaql_expression.YaqlExpression(string, self._version)
            yaql_expr(None)

        self.assertEqual(expected_calls, mock_parse.mock_calls)

    def test_is_expression_returns(self):
        expr = yaql_expression.YaqlExpression('string', self._version)

        with mock.patch('murano.dsl.yaql_integration.parse'):
            self.assertTrue(expr.is_expression('$some', self._version))
            self.assertTrue(expr.is_expression('$.someMore', self._version))

        with mock.patch('murano.dsl.yaql_integration.parse') as parse_mock:
            parse_mock.side_effect = exceptions.YaqlGrammarException
            self.assertFalse(expr.is_expression('', self._version))

        with mock.patch('murano.dsl.yaql_integration.parse') as parse_mock:
            parse_mock.side_effect = exceptions.YaqlLexicalException
            self.assertFalse(expr.is_expression('', self._version))

    def test_property(self):
        self.assertRaises(TypeError,
                          yaql_expression.YaqlExpression,
                          None, self._version)

        expr = yaql_expression.YaqlExpression('string', self._version)
        self.assertEqual(expr._version, expr.version)
        self.assertIsNone(expr._file_position)

        yaql_rep = expr.__repr__()
        self.assertEqual('YAQL(%s)' % expr._expression, yaql_rep)
